from typing import Optional
from jinja2 import Template


class ObservationSettings:
    """Time settings of the observation data. Used in feature join.
    Attributes:
        observation_path: path to the observation dataset, i.e. input dataset to get with features
        event_timestamp_column (Optional[str]): The timestamp field of your record. As sliding window aggregation feature assume each record in the source data should have a timestamp column. 
        timestamp_format (Optional[str], optional): The format of the timestamp field. Defaults to "epoch". Possible values are: 
        - `epoch` (seconds since epoch), for example `1647737463`
        - `epoch_millis` (milliseconds since epoch), for example `1647737517761`
        - Any date formats supported by [SimpleDateFormat](https://docs.oracle.com/javase/8/docs/api/java/text/SimpleDateFormat.html). 
            
    """
    def __init__(self,
                 observation_path: str,
                 event_timestamp_column: Optional[str] = None,
                 timestamp_format: str = "epoch") -> None:
        self.event_timestamp_column = event_timestamp_column
        self.timestamp_format = timestamp_format
        self.observation_path = observation_path

    def to_config(self) -> str:
        tm = Template("""
            {% if setting.event_timestamp_column is not none %}
            settings: {
                joinTimeSettings: {
                    timestampColumn: {
                        def: "{{setting.event_timestamp_column}}"
                        format: "{{setting.timestamp_format}}"
                    }
                }
            }
            {% endif %}
            observationPath: "{{setting.observation_path}}"
        """)
        return tm.render(setting=self)



from .client import *
from .feature_derivations import *
from .anchor import *
from .feature import *
from .source import *
from .dtype import *
from .source import *
from .transformation import *
from .transformation import *
from .typed_key import *
from .materialization_settings import (BackfillTime, MaterializationSettings)
from .sink import RedisSink
from .query_feature_list import FeatureQuery
from .lookup_feature import LookupFeature
from .aggregation import  Aggregation


import os
from datetime import datetime, timedelta
from pathlib import Path

from feathr.client import FeathrClient
from feathr import ValueType
from feathr.job_utils import get_result_df
from feathr import (BackfillTime, MaterializationSettings)
from feathr import FeatureQuery
from feathr import ObservationSettings
from feathr import RedisSink
from feathr import TypedKey
from test_fixture import snowflake_test_setup


def test_feathr_online_store_agg_features():
    """
    Test FeathrClient() get_online_features and batch_get can get feature data correctly.
    """
    test_workspace_dir = Path(__file__).parent.resolve() / "test_user_workspace"
    
    client = snowflake_test_setup(os.path.join(test_workspace_dir, "feathr_config.yaml"))

    backfill_time = BackfillTime(start=datetime(2020, 5, 20), end=datetime(2020, 5, 20), step=timedelta(days=1))
    redisSink = RedisSink(table_name="snowflakeSampleDemoFeature")
    settings = MaterializationSettings(name="snowflakeSampleDemoFeature",
                                   sinks=[redisSink],
                                   feature_names=['f_snowflake_call_center_division_name',
                                                  'f_snowflake_call_center_zipcode'],
                                   backfill_time=backfill_time)
    client.materialize_features(settings)
    # just assume the job is successful without validating the actual result in Redis. Might need to consolidate
    # this part with the test_feathr_online_store test case
    client.wait_job_to_finish(timeout_sec=600)

    res = client.get_online_features('snowflakeSampleDemoFeature', '1',
                                     ['f_snowflake_call_center_division_name', 'f_snowflake_call_center_zipcode'])

    assert len(res) == 2
    assert res[0] != None
    assert res[1] != None
    res = client.multi_get_online_features('snowflakeSampleDemoFeature',
                                    ['1', '2'],
                                    ['f_snowflake_call_center_division_name', 'f_snowflake_call_center_zipcode'])
    assert res['1'][0] != None
    assert res['1'][1] != None
    assert res['2'][0] != None
    assert res['2'][1] != None


def test_feathr_get_offline_features():
    """
    Test get_offline_features() can get feature data from Snowflake source correctly.
    """
    test_workspace_dir = Path(__file__).parent.resolve() / "test_user_workspace"
    
    
    client = snowflake_test_setup(os.path.join(test_workspace_dir, "feathr_config.yaml"))
    call_sk_id = TypedKey(key_column="CC_CALL_CENTER_SK",
                          key_column_type=ValueType.INT32,
                          description="call center sk",
                          full_name="snowflake.CC_CALL_CENTER_SK")

    feature_query = FeatureQuery(
        feature_list=['f_snowflake_call_center_division_name', 'f_snowflake_call_center_zipcode'],
        key=call_sk_id)
    settings = ObservationSettings(
        observation_path='jdbc:snowflake://dqllago-ol19457.snowflakecomputing.com/?user=feathrintegration&sfWarehouse'
                         '=COMPUTE_WH&dbtable=CALL_CENTER&sfDatabase=SNOWFLAKE_SAMPLE_DATA&sfSchema=TPCDS_SF10TCL')
    
    now = datetime.now()
     # set output folder based on different runtime
    if client.spark_runtime == 'databricks':
        output_path = ''.join(['dbfs:/feathrazure_cijob_snowflake','_', str(now.minute), '_', str(now.second), ".avro"])
    else:
        output_path = ''.join(['abfss://feathrazuretest3fs@feathrazuretest3storage.dfs.core.windows.net/demo_data/snowflake_output','_', str(now.minute), '_', str(now.second), ".avro"])
    
    client.get_offline_features(observation_settings=settings,
                                feature_query=feature_query,
                                output_path=output_path)

    # assuming the job can successfully run; otherwise it will throw exception
    client.wait_job_to_finish(timeout_sec=900)

    res = get_result_df(client)
    # just assume there are results.
    assert res.shape[0] > 1



import os
import re
import time
import urllib.request
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple
from urllib.parse import urlparse

from azure.identity import (ChainedTokenCredential, DefaultAzureCredential,
                            DeviceCodeCredential, EnvironmentCredential,
                            ManagedIdentityCredential)
from azure.storage.filedatalake import DataLakeServiceClient
from azure.synapse.spark import SparkClient
from azure.synapse.spark.models import (LivyStates, SparkBatchJob,
                                        SparkBatchJobOptions)
from loguru import logger
from tqdm import tqdm

from feathr._abc import SparkJobLauncher
from feathr.constants import *


class _FeathrSynapseJobLauncher(SparkJobLauncher):
    """
    Submits spark jobs to a Synapse spark cluster.
    """
    def __init__(self, synapse_dev_url: str, pool_name: str, datalake_dir: str, executor_size: str, executors: int):
        # use DeviceCodeCredential if EnvironmentCredential is not available
        self.credential = DefaultAzureCredential()
        # use the same credential for authentication to avoid further login.
        self._api = _SynapseJobRunner(
            synapse_dev_url, pool_name, executor_size=executor_size, executors=executors, credential=self.credential)
        self._datalake = _DataLakeFiler(
            datalake_dir, credential=self.credential)

    def upload_or_get_cloud_path(self, local_path_or_http_path: str):
        """
        Supports transferring file from an http path to cloud working storage, or upload directly from a local storage.
        """
        logger.info('Uploading {} to cloud..', local_path_or_http_path)
        res_path = self._datalake.upload_file_to_workdir(local_path_or_http_path)

        logger.info('{} is uploaded to location: {}', local_path_or_http_path, res_path)
        return res_path

    def download_result(self, result_path: str, local_folder: str):
        """
        Supports downloading files from the result folder
        """

        return self._datalake.download_file(result_path, local_folder)

    def submit_feathr_job(self, job_name: str, main_jar_path: str,  main_class_name: str, arguments: List[str],
                          reference_files_path: List[str], job_tags: Dict[str, str] = None,
                          configuration: Dict[str, str] = None):
        """
        Submits the feathr job
        Refer to the Apache Livy doc for more details on the meaning of the parameters:
        https://livy.apache.org/docs/latest/rest-api.html

        reference files: put everything there and the function will automatically categorize them based on the
        extension name to either the "files" argument in the Livy API, or the "jars" argument in the Livy API. The
        path can be local path and this function will automatically upload the function to the corresponding azure
        storage

        Also, note that the Spark application will automatically run on YARN cluster mode. You cannot change it if
        you are running with Azure Synapse.

        Args:
            job_name (str): name of the job
            main_jar_path (str): main file paths, usually your main jar file
            main_class_name (str): name of your main class
            arguments (str): all the arugments you want to pass into the spark job
            job_tags (str): tags of the job, for exmaple you might want to put your user ID, or a tag with a certain information
            configuration (Dict[str, str]): Additional configs for the spark job
        """

        if main_jar_path.startswith('abfs'):
            main_jar_cloud_path = main_jar_path
            logger.info(
                'Cloud path {} is used for running the job: {}', main_jar_path, job_name)
        else:
            logger.info('Uploading jar from {} to cloud for running job: {}',
                        main_jar_path, job_name)
            main_jar_cloud_path = self._datalake.upload_file_to_workdir(main_jar_path)
            logger.info('{} is uploaded to {} for running job: {}',
                         main_jar_path, main_jar_cloud_path, job_name)

        reference_file_paths = []
        for file_path in reference_files_path:
            reference_file_paths.append(
                self._datalake.upload_file_to_workdir(file_path))

        self.current_job_info = self._api.create_spark_batch_job(job_name=job_name,
                                                                 main_file=main_jar_cloud_path,
                                                                 class_name=main_class_name,
                                                                 arguments=arguments,
                                                                 reference_files=reference_files_path,
                                                                 tags=job_tags,
                                                                 configuration=configuration)
        logger.info('See submitted job here: https://web.azuresynapse.net/en-us/monitoring/sparkapplication')
        return self.current_job_info

    def wait_for_completion(self, timeout_seconds: Optional[float]) -> bool:
        """
        Returns true if the job completed successfully
        """
        start_time = time.time()
        while (timeout_seconds is None) or (time.time() - start_time < timeout_seconds):
            status = self.get_status()
            logger.info('Current Spark job status: {}', status)
            if status in {LivyStates.SUCCESS}:
                return True
            elif status in {LivyStates.ERROR, LivyStates.DEAD, LivyStates.KILLED}:
                return False
            else:
                time.sleep(30)
        else:
            raise TimeoutError('Timeout waiting for job to complete')

    def get_status(self) -> str:
        """Get current job status

        Returns:
            str: Status of the current job
        """
        job = self._api.get_spark_batch_job(self.current_job_info.id)
        assert job is not None
        return job.state

    def get_job_result_uri(self) -> str:
        """Get job output uri

        Returns:
            str: `output_path` field in the job tags
        """
        tags = self._api.get_spark_batch_job(self.current_job_info.id).tags
        assert tags is not None
        return tags[OUTPUT_PATH_TAG]


class _SynapseJobRunner(object):
    """
    Class to interact with Synapse Spark cluster
    """
    def __init__(self, synapse_dev_url, spark_pool_name, credential=None, executor_size='Small', executors=2):
        if credential is None:
            logger.warning('No valid Azure credential detected. Using DefaultAzureCredential')
            credential = DefaultAzureCredential()

        self.client = SparkClient(
            credential=credential,
            endpoint=synapse_dev_url,
            spark_pool_name=spark_pool_name
        )

        self._executor_size = executor_size
        self._executors = executors
        self.EXECUTOR_SIZE = {'Small': {'Cores': 4, 'Memory': '28g'}, 'Medium': {'Cores': 8, 'Memory': '56g'},
                              'Large': {'Cores': 16, 'Memory': '112g'}}

    def _categorized_files(self, reference_files: List[str]):
        """categorize files to make sure they are in the ready to submissio format

        Args:
            reference_files (List[str]): a list of reference files, can be either jars, pyfiles, etc.

        Returns:
            a tuple with all the files and jars clasified into list
        """
        if reference_files == None:
            return None, None

        files = []
        jars = []
        for file in reference_files:
            file = file.strip()
            if file.endswith('.jar'):
                jars.append(file)
            else:
                files.append(file)
        return files, jars

    def get_spark_batch_job(self, job_id:int):
        """
        Get the job object by searching a certain ID
        """

        return self.client.spark_batch.get_spark_batch_job(job_id, detailed=True)

    def get_spark_batch_jobs(self):
        """
        Get all the jobs in a certain Spark pool
        """

        return self.client.spark_batch.get_spark_batch_jobs(detailed=True)

    def cancel_spark_batch_job(self, job_id:int):
        """
        Cancel a job by searching a certain ID
        """

        return self.client.spark_batch.cancel_spark_batch_job(job_id)

    def create_spark_batch_job(self, job_name, main_file, class_name=None,
                               arguments=None,  reference_files=None, archives=None, configuration=None, tags=None):
        """
        Submit a spark job to a certain cluster
        """

        files, jars = self._categorized_files(reference_files)
        driver_cores = self.EXECUTOR_SIZE[self._executor_size]['Cores']
        driver_memory = self.EXECUTOR_SIZE[self._executor_size]['Memory']
        executor_cores = self.EXECUTOR_SIZE[self._executor_size]['Cores']
        executor_memory = self.EXECUTOR_SIZE[self._executor_size]['Memory']

        # Adding spaces between brackets. This is to workaround this known YARN issue (when running Spark on YARN):
        # https://issues.apache.org/jira/browse/SPARK-17814?focusedCommentId=15567964&page=com.atlassian.jira.plugin.system.issuetabpanels%3Acomment-tabpanel#comment-15567964
        # updated_arguments = []
        # for elem in arguments:
        #     if type(elem) == str:
        #         updated_arguments.append(elem.replace("}", " }"))
        #     else:
        #         updated_arguments.append(elem)

        spark_batch_job_options = SparkBatchJobOptions(
            tags=tags,
            name=job_name,
            file=main_file,
            class_name=class_name,
            arguments=arguments,
            jars=jars,
            files=files,
            archives=archives,
            configuration=configuration,
            driver_memory=driver_memory,
            driver_cores=driver_cores,
            executor_memory=executor_memory,
            executor_cores=executor_cores,
            executor_count=self._executors)

        return self.client.spark_batch.create_spark_batch_job(spark_batch_job_options, detailed=True)


class _DataLakeFiler(object):
    """
    Class to interact with Azure Data Lake Storage.
    """
    def __init__(self, datalake_dir, credential=None):
        # A datalake path would be something like this:
        # "abfss://feathrazuretest3fs@feathrazuretest3storage.dfs.core.windows.net/frame_getting_started" after this
        # split, datalake_path_split should give out something like this; ['abfss:', '', 'feathrazuretest3fs',
        # 'feathrazuretest3storage.dfs.core.windows.net', 'frame_getting_started'] datalake_path_split[0] should be
        # the protocal, datalake_path_split[1] is empty, datalake_path_split[2] is the file system used (aka
        # container name), datalake_path_split[3] should be the full name of this target path,
        # and datalake_path_split[3:] would be all the directory in this particular container split datalake names by
        # "/" or "@"
        datalake_path_split = list(filter(None, re.split('/|@', datalake_dir)))
        assert len(datalake_path_split) >= 3

        if credential is None:
            credential = DefaultAzureCredential()

        account_url = "https://" + datalake_path_split[2]

        self.file_system_client = DataLakeServiceClient(
            credential=credential,
            account_url=account_url
        ).get_file_system_client(datalake_path_split[1])

        if len(datalake_path_split) > 3:
            # directory exists in datalake path
            self.dir_client = self.file_system_client.get_directory_client(
                '/'.join(datalake_path_split[3:]))
            self.dir_client.create_directory()
        else:
            # otherwise use root folder instead
            self.dir_client = self.file_system_client.get_directory_client('/')

        self.datalake_dir = datalake_dir + \
            '/' if datalake_dir[-1] != '/' else datalake_dir

    def upload_file_to_workdir(self, src_file_path: str) -> str:
        """
        Handles file upload to the corresponding datalake storage. If a path starts with "wasb" or "abfs",
        it will skip uploading and return the original path; otherwise it will upload the source file to the working
        dir
        """

        src_parse_result = urlparse(src_file_path)
        if src_parse_result.scheme.startswith('http'):
            file_name = os.path.basename(src_file_path)
            file_client = self.dir_client.create_file(file_name)
            # returned paths for the uploaded file
            returned_path = self.datalake_dir + file_name
            with urllib.request.urlopen(src_file_path) as f:
                data = f.read()
                file_client.upload_data(data, overwrite=True)
                logger.info("{} is downloaded and then uploaded to location: {}", src_file_path, returned_path)
        elif src_parse_result.scheme.startswith('abfs') or src_parse_result.scheme.startswith('wasb'):
            # passed a cloud path
            logger.info("Skipping file {} as it's already in the cloud", src_file_path)
            returned_path = src_file_path
        else:
            # else it should be a local file path or dir
            if os.path.isdir(src_file_path):
                logger.info("Uploading folder {}", src_file_path)
                dest_paths = []
                for item in Path(src_file_path).glob('**/*.conf'):
                    returned_path = self.upload_file(item.resolve())
                    dest_paths.extend([returned_path])
                returned_path = ','.join(dest_paths)
            else:
                returned_path = self.upload_file(src_file_path)
        return returned_path

    def upload_file(self, src_file_path)-> str:
        file_name = os.path.basename(src_file_path)
        logger.info("Uploading file {}", file_name)
        file_client = self.dir_client.create_file(file_name)
        returned_path = self.datalake_dir + file_name
        with open(src_file_path, 'rb') as f:
            data = f.read()
            file_client.upload_data(data, overwrite=True)
        logger.info("{} is uploaded to location: {}", src_file_path, returned_path)
        return returned_path

    def download_file(self, target_adls_directory: str, local_dir_cache: str):
        """
        Download file to a local cache

        Args:
            target_adls_directory (str): target ADLS directory
            local_dir_cache (str): local cache to store local results
        """
        logger.info('Beginning reading of results from {}',
                    target_adls_directory)
        parse_result = urlparse(target_adls_directory)
        directory_client = self.file_system_client.get_directory_client(
            parse_result.path)
        # returns the paths to all the files in the target director in ADLS
        adls_paths = [file_path.name.split("/")[-1] for file_path in self.file_system_client.get_paths(
            path=parse_result.path) if not file_path.is_directory][1:]
        # need to generate list of local paths to write the files to
        local_paths = [os.path.join(local_dir_cache, file_name)
                       for file_name in adls_paths]
        for idx, file_to_write in enumerate(tqdm(adls_paths,desc="Downloading result files: ")):
            try:
                local_file = open(local_paths[idx], 'wb')
                file_client = directory_client.get_file_client(file_to_write)
                download = file_client.download_file()
                downloaded_bytes = download.readall()
                local_file.write(downloaded_bytes)
                local_file.close()
            except Exception as e:
                logger.error(e)
        logger.info('Finish downloading files from {} to {}.',
                    target_adls_directory,local_dir_cache)



from feathr.client import FeathrClient
import os
import glob
import pandavro as pdx
import pandas as pd
import tempfile


def get_result_df(client: FeathrClient) -> pd.DataFrame:
    """Download the job result dataset from cloud as a Pandas dataframe."""
    res_url = client.get_job_result_uri(block=True, timeout_sec=600)
    tmp_dir = tempfile.TemporaryDirectory()
    client.feathr_spark_laucher.download_result(result_path=res_url, local_folder=tmp_dir.name)
    dataframe_list = []
    # assuming the result are in avro format
    for file in glob.glob(os.path.join(tmp_dir.name, '*.avro')):
        dataframe_list.append(pdx.read_avro(file))
    vertical_concat_df = pd.concat(dataframe_list, axis=0)
    tmp_dir.cleanup()
    return vertical_concat_df


from feathr import FeatureAnchor
from feathr.client import FeathrClient
from feathr import BOOLEAN, FLOAT, INT32, STRING, ValueType
from feathr import Feature
from feathr import DerivedFeature
from feathr import INPUT_CONTEXT, HdfsSource
from feathr import WindowAggTransformation
from feathr import TypedKey
from datetime import datetime, timedelta
import os

def basic_test_setup(config_path: str):

    now = datetime.now()
    # set workspace folder by time; make sure we don't have write conflict if there are many CI tests running
    os.environ['SPARK_CONFIG__DATABRICKS__WORK_DIR'] = ''.join(['dbfs:/feathrazure_cijob','_', str(now.minute), '_', str(now.second), '_', str(now.microsecond)]) 
    os.environ['SPARK_CONFIG__AZURE_SYNAPSE__WORKSPACE_DIR'] = ''.join(['abfss://feathrazuretest3fs@feathrazuretest3storage.dfs.core.windows.net/feathr_github_ci','_', str(now.minute), '_', str(now.second) ,'_', str(now.microsecond)]) 
    
    client = FeathrClient(config_path=config_path)
    batch_source = HdfsSource(name="nycTaxiBatchSource",
                              path="abfss://feathrazuretest3fs@feathrazuretest3storage.dfs.core.windows.net/demo_data/green_tripdata_2020-04.csv",
                              event_timestamp_column="lpep_dropoff_datetime",
                              timestamp_format="yyyy-MM-dd HH:mm:ss")

    f_trip_distance = Feature(name="f_trip_distance",
                              feature_type=FLOAT, transform="trip_distance")
    f_trip_time_duration = Feature(name="f_trip_time_duration",
                                   feature_type=INT32,
                                   transform="time_duration(lpep_pickup_datetime, lpep_dropoff_datetime, 'minutes')")

    features = [
        f_trip_distance,
        f_trip_time_duration,
        Feature(name="f_is_long_trip_distance",
                feature_type=BOOLEAN,
                transform="cast_float(trip_distance)>30"),
        Feature(name="f_day_of_week",
                feature_type=INT32,
                transform="dayofweek(lpep_dropoff_datetime)"),
    ]


    request_anchor = FeatureAnchor(name="request_features",
                                   source=INPUT_CONTEXT,
                                   features=features)

    f_trip_time_distance = DerivedFeature(name="f_trip_time_distance",
                                          feature_type=FLOAT,
                                          input_features=[
                                              f_trip_distance, f_trip_time_duration],
                                          transform="f_trip_distance * f_trip_time_duration")

    f_trip_time_rounded = DerivedFeature(name="f_trip_time_rounded",
                                         feature_type=INT32,
                                         input_features=[f_trip_time_duration],
                                         transform="f_trip_time_duration % 10")

    location_id = TypedKey(key_column="DOLocationID",
                           key_column_type=ValueType.INT32,
                           description="location id in NYC",
                           full_name="nyc_taxi.location_id")
    agg_features = [Feature(name="f_location_avg_fare",
                            key=location_id,
                            feature_type=FLOAT,
                            transform=WindowAggTransformation(agg_expr="cast_float(fare_amount)",
                                                              agg_func="AVG",
                                                              window="90d")),
                    Feature(name="f_location_max_fare",
                            key=location_id,
                            feature_type=FLOAT,
                            transform=WindowAggTransformation(agg_expr="cast_float(fare_amount)",
                                                              agg_func="MAX",
                                                              window="90d"))
                    ]

    agg_anchor = FeatureAnchor(name="aggregationFeatures",
                               source=batch_source,
                               features=agg_features)

    client.build_features(anchor_list=[agg_anchor, request_anchor], derived_feature_list=[
        f_trip_time_distance, f_trip_time_rounded])

    return client




def snowflake_test_setup(config_path: str):
    now = datetime.now()
    # set workspace folder by time; make sure we don't have write conflict if there are many CI tests running
    os.environ['SPARK_CONFIG__DATABRICKS__WORK_DIR'] = ''.join(['dbfs:/feathrazure_cijob_snowflake','_', str(now.minute), '_', str(now.second), '_', str(now.microsecond)]) 
    os.environ['SPARK_CONFIG__AZURE_SYNAPSE__WORKSPACE_DIR'] = ''.join(['abfss://feathrazuretest3fs@feathrazuretest3storage.dfs.core.windows.net/feathr_github_ci_snowflake','_', str(now.minute), '_', str(now.second), '_', str(now.microsecond)]) 

    client = FeathrClient(config_path=config_path)
    batch_source = HdfsSource(name="snowflakeSampleBatchSource",
                              path="jdbc:snowflake://dqllago-ol19457.snowflakecomputing.com/?user=feathrintegration&sfWarehouse=COMPUTE_WH&dbtable=CALL_CENTER&sfDatabase=SNOWFLAKE_SAMPLE_DATA&sfSchema=TPCDS_SF10TCL",
                              )
    call_sk_id = TypedKey(key_column="CC_CALL_CENTER_SK",
                          key_column_type=ValueType.INT32,
                          description="call center sk",
                          full_name="snowflake.CC_CALL_CENTER_SK")

    f_snowflake_call_center_division_name = Feature(name="f_snowflake_call_center_division_name",feature_type=STRING, transform="CC_DIVISION_NAME", key=call_sk_id)

    f_snowflake_call_center_zipcode = Feature(name="f_snowflake_call_center_zipcode",feature_type=STRING, transform="CC_ZIP", key=call_sk_id)



    features = [f_snowflake_call_center_division_name, f_snowflake_call_center_zipcode ]


    snowflakeFeatures = FeatureAnchor(name="snowflakeFeatures",
                                   source=batch_source,
                                   features=features)




    client.build_features(anchor_list=[snowflakeFeatures])

    return client



from feathr import Feature
from feathr import ExpressionTransformation
from feathr import DerivedFeature
from feathr import BOOLEAN, FLOAT, FLOAT_VECTOR, ValueType
from feathr import TypedKey
import pytest

def assert_config_equals(one, another):
    assert one.translate(str.maketrans('', '', ' \n\t\r')) == another.translate(str.maketrans('', '', ' \n\t\r'))

def test_single_key_derived_feature_to_config():
    """Single key derived feature config generation should work"""
    user_key = TypedKey(full_name="mockdata.user", key_column="user_id", key_column_type=ValueType.INT32, description="An user identifier")
    user_embedding = Feature(name="user_embedding", feature_type=FLOAT_VECTOR, key=user_key)

    # A derived feature
    derived_feature = DerivedFeature(name="user_embemdding_derived",
                                        feature_type=FLOAT,
                                        key=user_key,
                                        input_features=user_embedding,
                                        transform="if_else(user_embedding, user_embedding, [])")

    derived_feature_config = """
    user_embemdding_derived: {
        key: [user_id]
        inputs: {
            user_embedding: {key: [user_id], feature: user_embedding}
        }
        definition: "if_else(user_embedding, user_embedding, [])"
        type: {
            type: TENSOR
            tensorCategory: DENSE
            dimensionType: []
            valType: FLOAT
        }
    }"""
    assert_config_equals(derived_feature.to_feature_config(), derived_feature_config)

def test_multikey_derived_feature_to_config():
    """Multikey derived feature config generation should work"""
    user_key = TypedKey(full_name="mockdata.user", key_column="user_id", key_column_type=ValueType.INT32, description="An user identifier")
    item_key = TypedKey(full_name="mockdata.item", key_column="item_id", key_column_type=ValueType.INT32, description="An item identifier")

    user_embedding = Feature(name="user_embedding", feature_type=FLOAT_VECTOR, key=user_key)
    item_embedding = Feature(name="item_embedding", feature_type=FLOAT_VECTOR, key=item_key)

    # A derived feature
    user_item_similarity = DerivedFeature(name="user_item_similarity",
                                        feature_type=FLOAT,
                                        key=[user_key, item_key],
                                        input_features=[user_embedding, item_embedding],
                                        transform="similarity(user_embedding, item_embedding)")

    derived_feature_config = """
    user_item_similarity: {
        key: [user_id, item_id]
        inputs: {
            user_embedding: {key: [user_id], feature: user_embedding}
            item_embedding: {key: [item_id], feature: item_embedding}
        }
        definition: "similarity(user_embedding,item_embedding)"
        type: {
            type: TENSOR
            tensorCategory: DENSE
            dimensionType: []
            valType: FLOAT
        }
    }"""
    assert_config_equals(user_item_similarity.to_feature_config(), derived_feature_config)


def test_derived_feature_to_config_with_alias():
    # More complicated use case, viewer viewee aliasged user key
    # References the same key feature with different alias
    user_key = TypedKey(full_name="mockdata.user", key_column="user_id", key_column_type=ValueType.INT32, description="An user identifier")
    user_embedding = Feature(name="user_embedding", key=user_key, feature_type=FLOAT_VECTOR)
    viewer_viewee_distance = DerivedFeature(name="viewer_viewee_distance",
                        key=[user_key.as_key("viewer"), user_key.as_key("viewee")],
                        feature_type=FLOAT,
                        input_features=[user_embedding.with_key("viewer").as_feature("viewer_embedding"),
                                        user_embedding.with_key("viewee").as_feature("viewee_embedding")],
                        transform="distance(viewer_embedding, viewee_embedding)")

    expected_feature_config = """
    viewer_viewee_distance: {
        key: [viewer, viewee]
        inputs: {
            viewer_embedding: {key: [viewer], feature: user_embedding}
            viewee_embedding: {key: [viewee], feature: user_embedding}
        }
        definition: "distance(viewer_embedding, viewee_embedding)"
        type: {
            type: TENSOR
            tensorCategory: DENSE
            dimensionType: []
            valType: FLOAT
        }
    }"""
    assert_config_equals(viewer_viewee_distance.to_feature_config(), expected_feature_config)


def test_multi_key_derived_feature_to_config_with_alias():
    # References the same relation feature key alias with different alias
    # Note that in this case, it is possible that distance(a, b) != distance(b,a)
    user_key = TypedKey(full_name="mockdata.user", key_column="user_id", key_column_type=ValueType.INT32, description="An user identifier")
    user_embedding = Feature(name="user_embedding", key=user_key, feature_type=FLOAT_VECTOR)

    viewer_viewee_distance = DerivedFeature(name="viewer_viewee_distance",
                        key=[user_key.as_key("viewer"), user_key.as_key("viewee")],
                        feature_type=FLOAT,
                        input_features=[user_embedding.with_key("viewer").as_feature("viewer_embedding"),
                                        user_embedding.with_key("viewee").as_feature("viewee_embedding")],
                        transform="distance(viewer_embedding, viewee_embedding)")

    viewee_viewer_combined = DerivedFeature(name = "viewee_viewer_combined_distance",
                                        key=[user_key.as_key("viewer"), user_key.as_key("viewee")],
                                        feature_type=FLOAT,
                                        input_features=[viewer_viewee_distance.with_key(["viewer", "viewee"])
                                                                            .as_feature("viewer_viewee_distance"),
                                                        viewer_viewee_distance.with_key(["viewee", "viewer"])
                                                                            .as_feature("viewee_viewer_distance"),],
                                        transform=ExpressionTransformation("viewer_viewee_distance + viewee_viewer_distance"))
    # Note that unlike key features, a relation feature does not need a feature anchor.

    expected_feature_config = """
    viewee_viewer_combined_distance: {
        key: [viewer, viewee]
        inputs: {
            viewer_viewee_distance: {key: [viewer, viewee], feature: viewer_viewee_distance}
            viewee_viewer_distance: {key: [viewee, viewer], feature: viewer_viewee_distance}
        }
        definition: "viewer_viewee_distance + viewee_viewer_distance"
        type: {
            type: TENSOR
            tensorCategory: DENSE
            dimensionType: []
            valType: FLOAT
        }
    }"""
    assert_config_equals(viewee_viewer_combined.to_feature_config(), expected_feature_config)

def test_derived_feature_on_multikey_anchored_feature_to_config():
    """Multikey derived feature config generation should work"""
    user_key = TypedKey(full_name="mockdata.user", key_column="user_id", key_column_type=ValueType.INT32, description="First part of an user identifier")
    user_key2 = TypedKey(full_name="mockdata.user2", key_column="user_id2", key_column_type=ValueType.INT32, description="Second part of an user identifier")

    user_embedding = Feature(name="user_embedding", feature_type=FLOAT_VECTOR, key=[user_key, user_key2])

    # A derived feature
    user_item_derived = DerivedFeature(name="user_item_similarity",
                                        feature_type=FLOAT,
                                        key=[user_key.as_key("viewer"), user_key2.as_key("viewee")],
                                        input_features=user_embedding.with_key(["viewer", "viewee"]),
                                        transform="if_else(user_embedding, user_embedding, [])")

    derived_feature_config = """
    user_item_similarity: {
        key: [viewer, viewee]
        inputs: {
            user_embedding: {key: [viewer, viewee], feature: user_embedding}
        }
        definition: "if_else(user_embedding, user_embedding, [])"
        type: {
            type: TENSOR
            tensorCategory: DENSE
            dimensionType: []
            valType: FLOAT
        }
    }"""
    assert_config_equals(user_item_derived.to_feature_config(), derived_feature_config)



def test_multi_key_derived_feature_to_config_with_wrong_alias():
    # References the same relation feature key alias with wrong alias
    # Should throw exception
    user_key = TypedKey(full_name="mockdata.user", key_column="user_id", key_column_type=ValueType.INT32, description="An user identifier")
    user_embedding = Feature(name="user_embedding", key=user_key, feature_type=FLOAT_VECTOR)

    with pytest.raises(AssertionError):
        viewer_viewee_distance = DerivedFeature(name="viewer_viewee_distance",
                                    key=[user_key.as_key("non_exist_alias"), user_key.as_key("viewee")],
                                    feature_type=FLOAT,
                                    input_features=[user_embedding.with_key("viewer").as_feature("viewer_embedding"),
                                                    user_embedding.with_key("viewee").as_feature("viewee_embedding")],
                                    transform="distance(viewer_embedding, viewee_embedding)")




from datetime import datetime, timedelta

from feathr._materialization_utils import _to_materialization_config
from feathr import (BackfillTime, MaterializationSettings)
from feathr import RedisSink


def test_feature_materialization_config():
    backfill_time = BackfillTime(start=datetime(2020, 5, 20), end=datetime(2020, 5,20), step=timedelta(days=1))
    redisSink = RedisSink(table_name="nycTaxiDemoFeature")
    settings = MaterializationSettings("nycTaxiTable",
                                        sinks=[redisSink],
                                        feature_names=["f_location_avg_fare", "f_location_max_fare"],
                                        backfill_time=backfill_time)
    config = _to_materialization_config(settings)
    expected_config = """ 
        operational: {
            name: nycTaxiTable
            endTime: "2020-05-20 00:00:00"
            endTimeFormat: "yyyy-MM-dd HH:mm:ss"
            resolution: DAILY
            output:[
                {
                    name: REDIS
                    params: {
                        table_name: "nycTaxiDemoFeature"
                    }
                }
            ]
        }
        features: [f_location_avg_fare, f_location_max_fare]
        """
    assert ''.join(config.split()) == ''.join(expected_config.split())


def test_feature_materialization_daily_schedule():
    """Test back fill cutoff time for a daily range"""
    backfill_time = BackfillTime(start=datetime(2022, 3, 1), end=datetime(2022, 3, 5), step=timedelta(days=1))
    settings = MaterializationSettings("", [], [], backfill_time)
    expected = [datetime(2022, 3, day) for day in range(1, 6)]
    assert settings.get_backfill_cutoff_time() == expected


def test_feature_materialization_hourly_schedule():
    """Test back fill cutoff time for a hourly range"""
    backfill_time = BackfillTime(start=datetime(2022, 3, 1, 1), end=datetime(2022, 3, 1, 5), step=timedelta(hours=1))
    settings = MaterializationSettings("", [], [], backfill_time)
    expected = [datetime(2022,3, 1, hour) for hour in range(1, 6)]
    assert settings.get_backfill_cutoff_time() == expected


def test_feature_materialization_now_schedule():
    """Test back fill cutoff time without backfill."""
    settings = MaterializationSettings("", [], [])
    date = settings.get_backfill_cutoff_time()[0]
    expected = datetime.now()
    assert expected.year == date.year
    assert expected.month == date.month
    assert expected.day == date.day



from enum import Enum


class Aggregation(Enum):
    """
    The built-in aggregation functions for LookupFeature
    """
    # No operation
    NOP = 0
    # Average
    AVG = 1
    MAX = 2
    MIN = 3
    SUM = 4
    UNION = 5
    # Element-wise average, typically used in array type value, i.e. 1d dense tensor
    ELEMENTWISE_AVG = 6
    ELEMENTWISE_MIN = 7
    ELEMENTWISE_MAX = 8
    ELEMENTWISE_SUM = 9
    # Pick the latest value according to its timestamp
    LATEST = 10


from abc import ABC, abstractmethod

from typing import Any, Dict, List, Optional, Tuple

class SparkJobLauncher(ABC):
    """This is the abstract class for all the spark launchers. All the Spark launcher should implement those interfaces
    """

    @abstractmethod
    def upload_or_get_cloud_path(self, local_path_or_http_path: str):
        """upload a file from local path or an http path to the current work directory. Should support transferring file from an http path to cloud working storage, or upload directly from a local storage.
        
        Args:
            local_path_or_http_path (str): local path or http path
        """
        pass


    @abstractmethod
    def submit_feathr_job(self, job_name: str, main_jar_path: str,  main_class_name: str, arguments: List[str],
                          reference_files_path: List[str], job_tags: Dict[str, str] = None,
                          configuration: Dict[str, str] = None):
        """
        Submits the feathr job

        Args:
            job_name (str): name of the job
            main_jar_path (str): main file paths, usually your main jar file
            main_class_name (str): name of your main class
            arguments (str): all the arugments you want to pass into the spark job
            job_tags (str): tags of the job, for exmaple you might want to put your user ID, or a tag with a certain information
            configuration (Dict[str, str]): Additional configs for the spark job
        """
        pass
    @abstractmethod
    def wait_for_completion(self, timeout_seconds: Optional[float]) -> bool:
        """Returns true if the job completed successfully

        Args:
            timeout_seconds (Optional[float]): time out secs

        Returns:
            bool: Returns true if the job completed successfully, otherwise False
        """
        pass

    @abstractmethod
    def get_status(self) -> str:
        """
        Get current job status

        Returns:
            str: Status of the current job

        Returns:
            str: _description_
        """
        pass




from abc import ABC, abstractmethod
from copy import deepcopy
from typing import List, Optional, Union

from jinja2 import Template

from feathr.dtype import FeatureType
from feathr.transformation import ExpressionTransformation, Transformation
from feathr.typed_key import DUMMY_KEY, TypedKey


class FeatureBase(ABC):
    """The base class for features
    It has a feature name, feature type, and a convenient transformation used to produce its feature value.
    Attributes:
        name: Unique name of the feature. Only alphabet, numbers, and '_' are allowed in the name.
                It can not start with numbers. Note that '.' is NOT ALLOWED!
        feature_type: the feature value type. e.g. INT32, FLOAT, etc. feathr.dtype
        key: The key of this feature. e.g. user_id.
        transform: A transformation used to produce its feature value. e.g. amount * 10
    """
    def __init__(self,
                 name: str,
                 feature_type: FeatureType,
                 transform: Optional[Union[str, Transformation]] = None,
                 key: Optional[Union[TypedKey, List[TypedKey]]] = [DUMMY_KEY]):
        self.name = name
        self.feature_type = feature_type
        self.key = key if isinstance(key, List) else [key]
        # feature_alias: Rename the derived feature to `feature_alias`. Default to feature name.
        self.feature_alias = name
        # If no transformation is specified, default to referencing the a field with the same name
        if transform is None:
            self.transform = ExpressionTransformation(name)
        elif isinstance(transform, str):
            self.transform = ExpressionTransformation(transform)
        else:
            self.transform = transform
        # An alias for the key in this feature. Default to its key column alias. Useful in derived features.
        self.key_alias = [k.key_column_alias for k in self.key]

    def with_key(self, key_alias: Union[str, List[str]]):
        """Rename the feature key with the alias. This is useful in derived features that depends on
        the same feature with different keys."""
        cleaned_key_alias = [key_alias] if isinstance(key_alias, str) else key_alias
        assert(len(cleaned_key_alias) == len(self.key))
        new_key = []
        for i in range(0, len(cleaned_key_alias)):
            typed_key = deepcopy(self.key[i])
            typed_key.key_column_alias = cleaned_key_alias[i]
            new_key.append(typed_key)

        res = deepcopy(self)
        res.key = new_key
        res.key_alias = cleaned_key_alias
        return res

    def as_feature(self, feature_alias):
        """Provide the feature a different alias, which can be used to reference the feature in transformation
        expression. This is useful in derived features that depends on the same feature with different keys."""
        new_feature = deepcopy(self)
        new_feature.feature_alias = feature_alias
        return new_feature

    @abstractmethod
    def to_feature_config(self) -> str:
        pass


class Feature(FeatureBase):
    """A feature is an individual measurable property or characteristic of an entity.
    It has a feature name, feature type, and a convenient row transformation used to produce its feature value.
    Attributes:
        name: Unique name of the feature. Only alphabet, numbers, and '_' are allowed in the name.
                It can not start with numbers. Note that '.' is NOT ALLOWED!
        feature_type: the feature value type. e.g. INT32, FLOAT, etc. Should be part of `feathr.dtype`
        key: The key of this feature. e.g. user_id.
        transform: A row transformation used to produce its feature value. e.g. amount * 10
    """
    def __init__(self,
                name: str,
                feature_type: FeatureType,
                key: Optional[Union[TypedKey, List[TypedKey]]] = [DUMMY_KEY],
                transform: Optional[Union[str, Transformation]] = None):
        super(Feature, self).__init__(name, feature_type, transform, key)

    def to_feature_config(self) -> str:
        tm = Template("""
            {{feature.name}}: {
                def:{{feature.transform.to_feature_config()}}
                {{feature.feature_type.to_feature_config()}} 
            }
        """)
        return tm.render(feature=self)




import base64
import logging
import math
import os
from datetime import datetime, timedelta
from pathlib import Path
from typing import List, Optional, Union
import tempfile

import redis
from jinja2 import Template
from pyhocon import ConfigFactory

from feathr._databricks_submission import _FeathrDatabricksJobLauncher
from feathr._envvariableutil import _EnvVaraibleUtil
from feathr._feature_registry import _FeatureRegistry
from feathr._file_utils import write_to_file
from feathr._materialization_utils import _to_materialization_config
from feathr._synapse_submission import _FeathrSynapseJobLauncher
from feathr.constants import *
from feathr.materialization_settings import MaterializationSettings
from feathr.protobuf.featureValue_pb2 import FeatureValue
from feathr.query_feature_list import FeatureQuery
from feathr.settings import ObservationSettings
from feathr.feature_derivations import DerivedFeature


class FeatureJoinJobParams:
    """Parameters related to feature join job.

    Attributes:
        join_config_path: Path to the join config.
        observation_path: Absolute path in Cloud to the observation data path.
        feature_config: Path to the features config.
        job_output_path: Absolute path in Cloud that you want your output data to be in.
    """

    def __init__(self, join_config_path, observation_path, feature_config, job_output_path):
        self.join_config_path = join_config_path
        self.observation_path = observation_path
        self.feature_config = feature_config
        self.job_output_path = job_output_path


class FeatureGenerationJobParams:
    """Parameters related to feature generation job.

    Attributes:
        generation_config_path: Path to the feature generation config.
        feature_config: Path to the features config.
    """

    def __init__(self, generation_config_path, feature_config):
        self.generation_config_path = generation_config_path
        self.feature_config = feature_config



class FeathrClient(object):
    """Feathr client.

    The client is used to create training dataset, materialize features, register features, and fetch features from
    the online storage.

    For offline storage and compute engine, Azure ADLS, AWS S3 and Azure Synapse are supported.

    For online storage, currently only Redis is supported.
    The users of this client is responsible for set up all the necessary information needed to start a Redis client via
    environment variable or a Spark cluster. Host address, port and password are needed to start the Redis client.

    Raises:
        RuntimeError: Fail to create the client since necessary environment variables are not set for Redis
        client creation.
    """

    def __init__(self, config_path = "./feathr_config.yaml", local_workspace_dir = None):

        self.logger = logging.getLogger(__name__)
        # Redis key separator
        self._KEY_SEPARATOR = ':'
        envutils = _EnvVaraibleUtil(config_path)
        if local_workspace_dir:
            self.local_workspace_dir = local_workspace_dir
        else:
            self.local_workspace_dir = tempfile.TemporaryDirectory().name
        
        self.envutils = envutils

        if not os.path.exists(config_path):
            self.logger.warning('Configuration path does not exist, you need to set the environment variables explicitly. For all the environment variables, please refer to https://github.com/linkedin/feathr/blob/main/feathr_project/feathrcli/data/feathr_user_workspace/feathr_config.yaml')

        # Load all configs from yaml at initialization
        # DO NOT load any configs from yaml during runtime.
        self.project_name = envutils.get_environment_variable_with_default(
            'project_config', 'project_name')

        # Redis configs
        self.redis_host = envutils.get_environment_variable_with_default(
            'online_store', 'redis', 'host')
        self.redis_port = envutils.get_environment_variable_with_default(
            'online_store', 'redis', 'port')
        self.redis_ssl_enabled = envutils.get_environment_variable_with_default(
            'online_store', 'redis', 'ssl_enabled')

        # S3 configs
        self.s3_endpoint = envutils.get_environment_variable_with_default(
            'offline_store', 's3', 's3_endpoint')

        # spark configs
        self.output_num_parts = envutils.get_environment_variable_with_default(
            'spark_config', 'spark_result_output_parts')
        self.spark_runtime = envutils.get_environment_variable_with_default(
            'spark_config', 'spark_cluster')

        if self.spark_runtime not in {'azure_synapse', 'databricks'}:
            raise RuntimeError(
                'Only \'azure_synapse\' and \'databricks\' are currently supported.')
        elif self.spark_runtime == 'azure_synapse':
            # Feathr is a spark-based application so the feathr jar compiled from source code will be used in the
            # Spark job submission. The feathr jar hosted in cloud saves the time users needed to upload the jar from
            # their local env.
            self._FEATHR_JOB_JAR_PATH = \
                envutils.get_environment_variable_with_default(
                    'spark_config', 'azure_synapse', 'feathr_runtime_location')

            self.feathr_spark_laucher = _FeathrSynapseJobLauncher(
                synapse_dev_url=envutils.get_environment_variable_with_default(
                    'spark_config', 'azure_synapse', 'dev_url'),
                pool_name=envutils.get_environment_variable_with_default(
                    'spark_config', 'azure_synapse', 'pool_name'),
                datalake_dir=envutils.get_environment_variable_with_default(
                    'spark_config', 'azure_synapse', 'workspace_dir'),
                executor_size=envutils.get_environment_variable_with_default(
                    'spark_config', 'azure_synapse', 'executor_size'),
                executors=envutils.get_environment_variable_with_default(
                    'spark_config', 'azure_synapse', 'executor_num')
            )
        elif self.spark_runtime == 'databricks':
            # Feathr is a spark-based application so the feathr jar compiled from source code will be used in the
            # Spark job submission. The feathr jar hosted in cloud saves the time users needed to upload the jar from
            # their local env.
            self._FEATHR_JOB_JAR_PATH = \
                envutils.get_environment_variable_with_default(
                    'spark_config', 'databricks', 'feathr_runtime_location')

            self.feathr_spark_laucher = _FeathrDatabricksJobLauncher(
                workspace_instance_url=envutils.get_environment_variable_with_default(
                    'spark_config', 'databricks', 'workspace_instance_url'),
                token_value=_EnvVaraibleUtil.get_environment_variable(
                    'DATABRICKS_WORKSPACE_TOKEN_VALUE'),
                config_template=envutils.get_environment_variable_with_default(
                    'spark_config', 'databricks', 'config_template'),
                databricks_work_dir=envutils.get_environment_variable_with_default(
                    'spark_config', 'databricks', 'work_dir')
            )

        self._construct_redis_client()

        self.registry = _FeatureRegistry(config_path = config_path)

    def _check_required_environment_variables_exist(self):
        """Checks if the required environment variables(form feathr_config.yaml) is set.

        Some required information has to be set via environment variables so the client can work.
        """
        for required_field in self.required_fields:
            if required_field not in os.environ:
                raise RuntimeError(f'{required_field} is not set in environment variable. All required environment '
                                   f'variables are: {self.required_fields}.')

    def register_features(self):
        """Registers features based on the current workspace
        """
        if 'anchor_list' in dir(self) and 'derived_feature_list' in dir(self):
            _FeatureRegistry.save_to_feature_config_from_context(self.anchor_list, self.derived_feature_list, self.local_workspace_dir)
        else:
            RuntimeError("Please call FeathrClient.build_features() first in order to register features")
        self.registry.register_features(self.local_workspace_dir)
    
    def build_features(self, anchor_list, derived_feature_list: Optional[List[DerivedFeature]] = []):
        """Registers features based on the current workspace
        """
        self.registry.save_to_feature_config_from_context(anchor_list, derived_feature_list, self.local_workspace_dir)
        self.anchor_list = anchor_list
        self.derived_feature_list = derived_feature_list

    def list_registered_features(self, project_name: str = None) -> List[str]:
        """List all the already registered features. If project_name is not provided or is None, it will return all
        the registered features; otherwise it will only return features under this project
        """
        return self.registry.list_registered_features(project_name)

    def get_registry_client(self):
        """
        Returns registry client in case users want to perform more advanced operations
        """
        return self.registry.get_registry_client()

    def get_online_features(self, feature_table, key, feature_names):
        """Fetches feature value for a certain key from a online feature table.

        Args:
            feature_table: the name of the feature table.
            key: the key of the entity
            feature_names: list of feature names to fetch

        Return:
            A list of feature values for this entity. It's ordered by the requested feature names.
            For example, feature_names = ['f_is_medium_trip_distance', 'f_day_of_week', 'f_day_of_month', 'f_hour_of_day']
            then, the returned feature values is: [b'true', b'4.0', b'31.0', b'23.0'].
            If the feature_table or key doesn't exist, then a list of Nones are returned. For example,
            [None, None, None, None].
            If a feature doesn't exist, then a None is returned for that feature. For example:
            [None, b'4.0', b'31.0', b'23.0'].
            """
        redis_key = self._construct_redis_key(feature_table, key)
        res = self.redis_clint.hmget(redis_key, *feature_names)
        return self._decode_proto(res)

    def multi_get_online_features(self, feature_table, keys, feature_names):
        """Fetches feature value for a list of keys from a online feature table. This is the batch version of the get API.

        Args:
            feature_table: the name of the feature table.
            keys: list of keys for the entities
            feature_names: list of feature names to fetch

        Return:
            A list of feature values for the requested entities. It's ordered by the requested feature names. For
            example, keys = [12, 24], feature_names = ['f_is_medium_trip_distance', 'f_day_of_week', 'f_day_of_month',
            'f_hour_of_day'] then, the returned feature values is: {'12': [b'false', b'5.0', b'1.0', b'0.0'],
            '24': [b'true', b'4.0', b'31.0', b'23.0']}. If the feature_table or key doesn't exist, then a list of Nones
            are returned. For example, {'12': [None, None, None, None], '24': [None, None, None, None]} If a feature
            doesn't exist, then a None is returned for that feature. For example: {'12': [None, b'4.0', b'31.0',
            b'23.0'], '24': [b'true', b'4.0', b'31.0', b'23.0']}.
        """
        with self.redis_clint.pipeline() as redis_pipeline:
            for key in keys:
                redis_key = self._construct_redis_key(feature_table, key)
                redis_pipeline.hmget(redis_key, *feature_names)
            pipeline_result = redis_pipeline.execute()

        decoded_pipeline_result = []
        for feature_list in pipeline_result:
            decoded_pipeline_result.append(self._decode_proto(feature_list))

        return dict(zip(keys, decoded_pipeline_result))

    def _decode_proto(self, feature_list):
        """Decode the bytes(in string form) via base64 decoder. For dense array, it will be returned as Python List.
        For sparse array, it will be returned as tuple of index array and value array. The order of elements in the
        arrays won't be changed.
        """
        typed_result = []
        for raw_feature in feature_list:
            if raw_feature:
                feature_value = FeatureValue()
                decoded = base64.b64decode(raw_feature)
                feature_value.ParseFromString(decoded)
                if feature_value.WhichOneof('FeatureValueOneOf') == 'boolean_value':
                    typed_result.append(feature_value.boolean_value)
                elif feature_value.WhichOneof('FeatureValueOneOf') == 'string_value':
                    typed_result.append(feature_value.string_value)
                elif feature_value.WhichOneof('FeatureValueOneOf') == 'float_value':
                    typed_result.append(feature_value.float_value)
                elif feature_value.WhichOneof('FeatureValueOneOf') == 'double_value':
                    typed_result.append(feature_value.double_value)
                elif feature_value.WhichOneof('FeatureValueOneOf') == 'int_value':
                    typed_result.append(feature_value.int_value)
                elif feature_value.WhichOneof('FeatureValueOneOf') == 'long_value':
                    typed_result.append(feature_value.long_value)
                elif feature_value.WhichOneof('FeatureValueOneOf') == 'int_array':
                    typed_result.append(feature_value.int_array.integers)
                elif feature_value.WhichOneof('FeatureValueOneOf') == 'string_array':
                    typed_result.append(feature_value.string_array.strings)
                elif feature_value.WhichOneof('FeatureValueOneOf') == 'float_array':
                    typed_result.append(feature_value.float_array.floats)
                elif feature_value.WhichOneof('FeatureValueOneOf') == 'double_array':
                    typed_result.append(feature_value.double_array.doubles)
                elif feature_value.WhichOneof('FeatureValueOneOf') == 'boolean_array':
                    typed_result.append(feature_value.boolean_array.booleans)
                elif feature_value.WhichOneof('FeatureValueOneOf') == 'sparse_string_array':
                    typed_result.append((feature_value.sparse_string_array.index_integers, feature_value.sparse_string_array.value_strings))
                elif feature_value.WhichOneof('FeatureValueOneOf') == 'sparse_bool_array':
                    typed_result.append((feature_value.sparse_bool_array.index_integers, feature_value.sparse_bool_array.value_booleans))
                elif feature_value.WhichOneof('FeatureValueOneOf') == 'sparse_float_array':
                    typed_result.append((feature_value.sparse_float_array.index_integers, feature_value.sparse_float_array.value_floats))
                elif feature_value.WhichOneof('FeatureValueOneOf') == 'sparse_double_array':
                    typed_result.append((feature_value.sparse_double_array.index_integers, feature_value.sparse_double_array.value_doubles))
                elif feature_value.WhichOneof('FeatureValueOneOf') == 'sparse_long_array':
                    typed_result.append((feature_value.sparse_long_array.index_integers, feature_value.sparse_long_array.value_longs))
                else:
                    self.logger.debug("Fail to load the feature type. Maybe a new type that is not supported by this "
                                      "client version")
                    self.logger.debug(f"The raw feature is {raw_feature}.")
                    self.logger.debug(f"The loaded feature is {feature_value}")
                    typed_result.append(None)
            else:
                typed_result.append(raw_feature)
        return typed_result

    def _clean_test_data(self, feature_table):
        """
        WARNING: THIS IS ONLY USED FOR TESTING
        Clears a namespace in redis cache.
        This may be very time consuming.

        Args:
          feature_table: str, feature_table i.e your prefix before the separator in the Redis database.
        """
        cursor = '0'
        ns_keys = feature_table + '*'
        while cursor != 0:
            # 5000 count at a scan seems reasonable faster for our testing data
            cursor, keys = self.redis_clint.scan(
                cursor=cursor, match=ns_keys, count=5000)
            if keys:
                self.redis_clint.delete(*keys)

    def _construct_redis_key(self, feature_table, key):
        return feature_table + self._KEY_SEPARATOR + key

    def _construct_redis_client(self):
        """Constructs the Redis client. The host, port, credential and other parameters can be set via environment
        parameters.
        """
        password = _EnvVaraibleUtil.get_environment_variable(REDIS_PASSWORD)
        host = self.redis_host
        port = self.redis_port
        ssl_enabled = self.redis_ssl_enabled

        redis_clint = redis.Redis(
            host=host,
            port=port,
            password=password,
            ssl=ssl_enabled)
        self.logger.info('Redis connection is successful and completed.')
        self.redis_clint = redis_clint


    def get_offline_features(self,
                             observation_settings: ObservationSettings,
                             feature_query: Union[FeatureQuery, List[FeatureQuery]],
                             output_path: str
                             ):
        """
        Get offline features for the observation dataset
        Args:
            observation_settings: settings of the observation data, e.g. timestamp columns, input path, etc.
            feature_query: features that are requested to add onto the observation data
            output_path: output path of job, i.e. the observation data with features attached.
        """
        # produce join config
        tm = Template("""
            {{observation_settings.to_config()}}
            featureList: [
                {% for list in feature_lists %}
                    {{list.to_config()}}
                {% endfor %}
            ]
            outputPath: "{{output_path}}"
        """)
        feature_queries = feature_query if isinstance(feature_query, List) else [feature_query]
        config = tm.render(feature_lists=feature_queries, observation_settings=observation_settings, output_path=output_path)
        config_file_name = "feature_join_conf/feature_join.conf"
        config_file_path = os.path.join(self.local_workspace_dir, config_file_name)

        # make sure `FeathrClient.build_features()` is called before getting offline features/materialize features
        # otherwise users will be confused on what are the available features
        # in build_features it will assign anchor_list and derived_feature_list variable, hence we are checking if those two variables exist to make sure the above condition is met
        if 'anchor_list' in dir(self) and 'derived_feature_list' in dir(self):
            _FeatureRegistry.save_to_feature_config_from_context(self.anchor_list, self.derived_feature_list, self.local_workspace_dir)
        else:
            RuntimeError("Please call FeathrClient.build_features() first in order to get offline features")

        write_to_file(content=config, full_file_name=config_file_path)
        return self._get_offline_features_with_config(config_file_path)

    def _get_offline_features_with_config(self, feature_join_conf_path='feature_join_conf/feature_join.conf'):
        """Joins the features to your offline observation dataset based on the join config.

        Args:
          feature_join_conf_path: Relative path to your feature join config file.
        """
        
            
        feathr_feature = ConfigFactory.parse_file(feature_join_conf_path)

        feature_join_job_params = FeatureJoinJobParams(join_config_path=os.path.abspath(feature_join_conf_path),
                                                       observation_path=feathr_feature['observationPath'],
                                                       feature_config=os.path.join(self.local_workspace_dir, 'feature_conf/'),
                                                       job_output_path=feathr_feature['outputPath'],
                                                       )

        # submit the jars
        return self.feathr_spark_laucher.submit_feathr_job(
            job_name=self.project_name + '_feathr_feature_join_job',
            main_jar_path=self._FEATHR_JOB_JAR_PATH,
            job_tags={OUTPUT_PATH_TAG:feature_join_job_params.job_output_path},
            main_class_name='com.linkedin.feathr.offline.job.FeatureJoinJob',
            arguments=[
                '--join-config', self.feathr_spark_laucher.upload_or_get_cloud_path(
                    feature_join_job_params.join_config_path),
                '--input', feature_join_job_params.observation_path,
                '--output', feature_join_job_params.job_output_path,
                '--feature-config', self.feathr_spark_laucher.upload_or_get_cloud_path(
                    feature_join_job_params.feature_config),
                '--num-parts', self.output_num_parts,
                '--s3-config', self._get_s3_config_str(),
                '--adls-config', self._get_adls_config_str(),
                '--blob-config', self._get_blob_config_str(),
                '--sql-config', self._get_sql_config_str(),
                '--snowflake-config', self._get_snowflake_config_str()
            ],
            reference_files_path=[],
        )

    def get_job_result_uri(self, block=True, timeout_sec=300):
        """Gets the job output URI
        """
        if not block:
            return self.feathr_spark_laucher.get_job_result_uri()
        # Block the API by pooling the job status and wait for complete
        if self.feathr_spark_laucher.wait_for_completion(timeout_sec):
            return self.feathr_spark_laucher.get_job_result_uri()
        else:
            raise RuntimeError(
                'Spark job failed so output cannot be retrieved.')

    def wait_job_to_finish(self, timeout_sec: int = 300):
        """Waits for the job to finish in a blocking way unless it times out
        """
        if self.feathr_spark_laucher.wait_for_completion(timeout_sec):
            return
        else:
            raise RuntimeError('Spark job failed.')

    def materialize_features(self, settings: MaterializationSettings):
        """Materialize feature data

        Args:
            settings: Feature materialization settings
        """
        # produce materialization config

        for end in settings.get_backfill_cutoff_time():
            settings.backfill_time.end = end
            config = _to_materialization_config(settings)
            config_file_name = "feature_gen_conf/auto_gen_config_{}.conf".format(end.timestamp())
            config_file_path = os.path.join(self.local_workspace_dir, config_file_name)
            write_to_file(content=config, full_file_name=config_file_path)
            
            # make sure `FeathrClient.build_features()` is called before getting offline features/materialize features in the python SDK
            # otherwise users will be confused on what are the available features
            # in build_features it will assign anchor_list and derived_feature_list variable, hence we are checking if those two variables exist to make sure the above condition is met
            if 'anchor_list' in dir(self) and 'derived_feature_list' in dir(self):
                _FeatureRegistry.save_to_feature_config_from_context(self.anchor_list, self.derived_feature_list, self.local_workspace_dir)
            else:
                RuntimeError("Please call FeathrClient.build_features() first in order to materialize the features")

            # CLI will directly call this so the experiene won't be broken
            self._materialize_features_with_config(config_file_path)
            if os.path.exists(config_file_path):
                os.remove(config_file_path)

    def _materialize_features_with_config(self, feature_gen_conf_path: str = 'feature_gen_conf/feature_gen.conf'):
        """Materializes feature data based on the feature generation config. The feature
        data will be materialized to the destination specified in the feature generation config.

        Args
          feature_gen_conf_path: Relative path to the feature generation config you want to materialize.
        """
        
        

        # Read all features conf
        generation_config = FeatureGenerationJobParams(
            generation_config_path=os.path.abspath(feature_gen_conf_path),
            feature_config=os.path.join(self.local_workspace_dir, "feature_conf/"))

        return self.feathr_spark_laucher.submit_feathr_job(
            job_name=self.project_name + '_feathr_feature_materialization_job',
            main_jar_path=self._FEATHR_JOB_JAR_PATH,
            main_class_name='com.linkedin.feathr.offline.job.FeatureGenJob',
            arguments=[
                '--generation-config', self.feathr_spark_laucher.upload_or_get_cloud_path(
                    generation_config.generation_config_path),
                # Local Config, comma seperated file names
                '--feature-config', self.feathr_spark_laucher.upload_or_get_cloud_path(
                    generation_config.feature_config),
                '--redis-config', self._getRedisConfigStr(),
                '--s3-config', self._get_s3_config_str(),
                '--adls-config', self._get_adls_config_str(),
                '--blob-config', self._get_blob_config_str(),
                '--sql-config', self._get_sql_config_str(),
                '--snowflake-config', self._get_snowflake_config_str()
            ],
            reference_files_path=[],
        )


    def wait_job_to_finish(self, timeout_sec: int = 300):
        """Waits for the job to finish in a blocking way unless it times out
        """
        if self.feathr_spark_laucher.wait_for_completion(timeout_sec):
            return
        else:
            raise RuntimeError('Spark job failed.')

    def _getRedisConfigStr(self):
        """Construct the Redis config string. The host, port, credential and other parameters can be set via environment
        variables."""
        password = _EnvVaraibleUtil.get_environment_variable(REDIS_PASSWORD)
        host = self.redis_host
        port = self.redis_port
        ssl_enabled = self.redis_ssl_enabled
        config_str = """
        REDIS_PASSWORD: "{REDIS_PASSWORD}"
        REDIS_HOST: "{REDIS_HOST}"
        REDIS_PORT: {REDIS_PORT}
        REDIS_SSL_ENABLED: {REDIS_SSL_ENABLED}
        """.format(REDIS_PASSWORD=password, REDIS_HOST=host, REDIS_PORT=port, REDIS_SSL_ENABLED=ssl_enabled)
        return config_str

    def _get_s3_config_str(self):
        """Construct the S3 config string. The endpoint, access key, secret key, and other parameters can be set via
        environment variables."""
        endpoint = self.s3_endpoint
        # if s3 endpoint is set in the feathr_config, then we need other environment variables
        # keys can't be only accessed through environment
        access_key = _EnvVaraibleUtil.get_environment_variable('S3_ACCESS_KEY')
        secret_key = _EnvVaraibleUtil.get_environment_variable('S3_SECRET_KEY')
        # HOCCON format will be parsed by the Feathr job
        config_str = """
            S3_ENDPOINT: {S3_ENDPOINT}
            S3_ACCESS_KEY: "{S3_ACCESS_KEY}"
            S3_SECRET_KEY: "{S3_SECRET_KEY}"
            """.format(S3_ENDPOINT=endpoint, S3_ACCESS_KEY=access_key, S3_SECRET_KEY=secret_key)
        return config_str

    def _get_adls_config_str(self):
        """Construct the ADLS config string for abfs(s). The Account, access key and other parameters can be set via
        environment variables."""
        account = _EnvVaraibleUtil.get_environment_variable('ADLS_ACCOUNT')
        # if ADLS Account is set in the feathr_config, then we need other environment variables
        # keys can't be only accessed through environment
        key = _EnvVaraibleUtil.get_environment_variable('ADLS_KEY')
        # HOCCON format will be parsed by the Feathr job
        config_str = """
            ADLS_ACCOUNT: {ADLS_ACCOUNT}
            ADLS_KEY: "{ADLS_KEY}"
            """.format(ADLS_ACCOUNT=account, ADLS_KEY=key)
        return config_str

    def _get_blob_config_str(self):
        """Construct the Blob config string for wasb(s). The Account, access key and other parameters can be set via
        environment variables."""
        account = _EnvVaraibleUtil.get_environment_variable('BLOB_ACCOUNT')
        # if BLOB Account is set in the feathr_config, then we need other environment variables
        # keys can't be only accessed through environment
        key = _EnvVaraibleUtil.get_environment_variable('BLOB_KEY')
        # HOCCON format will be parsed by the Feathr job
        config_str = """
            BLOB_ACCOUNT: {BLOB_ACCOUNT}
            BLOB_KEY: "{BLOB_KEY}"
            """.format(BLOB_ACCOUNT=account, BLOB_KEY=key)
        return config_str

    def _get_sql_config_str(self):
        """Construct the SQL config string for jdbc. The dbtable (query), user, password and other parameters can be set via
        environment variables."""
        table = _EnvVaraibleUtil.get_environment_variable('JDBC_TABLE')
        user = _EnvVaraibleUtil.get_environment_variable('JDBC_USER')
        password = _EnvVaraibleUtil.get_environment_variable('JDBC_PASSWORD')
        driver = _EnvVaraibleUtil.get_environment_variable('JDBC_DRIVER')
        auth_flag = _EnvVaraibleUtil.get_environment_variable('JDBC_AUTH_FLAG')
        token = _EnvVaraibleUtil.get_environment_variable('JDBC_TOKEN')
        # HOCCON format will be parsed by the Feathr job
        config_str = """
            JDBC_TABLE: {JDBC_TABLE}
            JDBC_USER: {JDBC_USER}
            JDBC_PASSWORD: {JDBC_PASSWORD}
            JDBC_DRIVER: {JDBC_DRIVER}
            JDBC_AUTH_FLAG: {JDBC_AUTH_FLAG}
            JDBC_TOKEN: {JDBC_TOKEN}
            """.format(JDBC_TABLE=table, JDBC_USER=user, JDBC_PASSWORD=password, JDBC_DRIVER = driver, JDBC_AUTH_FLAG = auth_flag, JDBC_TOKEN = token)
        return config_str

    def _get_snowflake_config_str(self):
        """Construct the Snowflake config string for jdbc. The url, user, role and other parameters can be set via
        yaml config. Password can be set via environment variables."""
        sf_url = self.envutils.get_environment_variable_with_default('offline_store', 'snowflake', 'url')
        sf_user = self.envutils.get_environment_variable_with_default('offline_store', 'snowflake', 'user')
        sf_role = self.envutils.get_environment_variable_with_default('offline_store', 'snowflake', 'role')
        sf_password = self.envutils.get_environment_variable('JDBC_SF_PASSWORD')
        # HOCCON format will be parsed by the Feathr job
        config_str = """
            JDBC_SF_URL: {JDBC_SF_URL}
            JDBC_SF_USER: {JDBC_SF_USER}
            JDBC_SF_ROLE: {JDBC_SF_ROLE}
            JDBC_SF_PASSWORD: {JDBC_SF_PASSWORD}
            """.format(JDBC_SF_URL=sf_url, JDBC_SF_USER=sf_user, JDBC_SF_PASSWORD=sf_password, JDBC_SF_ROLE=sf_role)
        return config_str

    def get_features_from_registry(self, project_name):
        """ Sync features from the registry given a project name """
        # TODO - Add support for customized workspace path
        self.registry.get_features_from_registry(project_name, os.path.abspath("./"))


import os
import yaml
from loguru import logger


class _EnvVaraibleUtil(object):
    def __init__(self, config_path):
        self.config_path = config_path

    def get_environment_variable_with_default(self, *args):
        """Gets the environment variable for the variable key.
        Args:
            *args: list of keys in feathr_config.yaml file
        Return:
            A environment variable for the variable key. If it's not set in the environment, then a default is retrieved
            from the feathr_config.yaml file with the same config key.
            """

        # if envs exist, just return the existing env variable without reading the file
        env_keyword = "__".join(args)
        upper_env_keyword = env_keyword.upper()
        # make it work for lower case and upper case.
        env_variable = os.environ.get(
            env_keyword, os.environ.get(upper_env_keyword))
        if env_variable:
            return env_variable

        # if the config path doesn't exist, just return
        try:
            assert os.path.exists(os.path.abspath(self.config_path))
        except:
            logger.info("{} is not set and configuration file {} cannot be found. One of those shoudl be set." , env_keyword, self.config_path)

        with open(os.path.abspath(self.config_path), 'r') as stream:
            try:
                yaml_config = yaml.safe_load(stream)
                # concat all layers
                # check in environment variable
                yaml_layer = yaml_config

                # resolve one layer after another
                for arg in args:
                    yaml_layer = yaml_layer[arg]
                return yaml_layer
            except yaml.YAMLError as exc:
                logger.info(exc)

    @staticmethod
    def get_environment_variable(variable_key):
        """Gets the environment variable for the variable key.

        Args:
            variable_key: environment variable key that is used to retrieve the environment variable
        Return:
            A environment variable for the variable key.
        Raises:
            ValueError: If the environment variable is not set for this key, an exception is thrown.
            """
        password = os.environ.get(variable_key)
        if not password:
            logger.info(variable_key +
                        ' is not set in the environment variables.')
        return password



import enum
from typing import Type, Union, List
from abc import ABC, abstractmethod

from jinja2 import Template

class ValueType(enum.Enum):
    """
    Basic value type. Used as feature key type.
    """
    UNSPECIFIED = 0
    BOOL = 1
    INT32 = 2
    INT64 = 3
    FLOAT = 4
    DOUBLE = 5
    STRING = 6
    BYTES = 7


class FeatureType(ABC):
    """Base class for all feature types"""
    @abstractmethod
    def to_feature_config(self) -> str:
        pass

class BooleanFeatureType(FeatureType):
    def to_feature_config(self) -> str:
        return """
           type: {
                type: TENSOR
                tensorCategory: DENSE
                dimensionType: []
                valType: BOOLEAN
            }
        """

class Int32FeatureType(FeatureType):
    def to_feature_config(self) -> str:
        return """
            type: {
                type: TENSOR
                tensorCategory: DENSE
                dimensionType: []
                valType: INT
            }
        """

class Int64FeatureType(FeatureType):
    def to_feature_config(self) -> str:
        return """
            type: {
                type: TENSOR
                tensorCategory: DENSE
                dimensionType: []
                valType: LONG
            }
        """

class FloatFeatureType(FeatureType):
    def to_feature_config(self) -> str:
        return """
            type: {
                type: TENSOR
                tensorCategory: DENSE
                dimensionType: []
                valType: FLOAT
            }
        """

class DoubleFeatureType(FeatureType):
    def to_feature_config(self) -> str:
        return """
            type: {
                type: TENSOR
                tensorCategory: DENSE
                dimensionType: []
                valType: DOUBLE
            }
        """

class StringFeatureType(FeatureType):
    def to_feature_config(self) -> str:
        return """
            type: {
                type: TENSOR
                tensorCategory: DENSE
                dimensionType: []
                valType: STRING
            }
        """

class BytesFeatureType(FeatureType):
    def to_feature_config(self) -> str:
        return """
            type: {
                type: TENSOR
                tensorCategory: DENSE
                dimensionType: []
                valType: BYTES
            }
        """

class FloatVectorFeatureType(FeatureType):
    def to_feature_config(self) -> str:
        return """
           type: {
                type: TENSOR
                tensorCategory: DENSE
                dimensionType: [INT]
                valType: FLOAT
            }
        """


class Int32VectorFeatureType(FeatureType):
    def to_feature_config(self) -> str:
        return """
           type: {
                type: TENSOR
                tensorCategory: DENSE
                dimensionType: [INT]
                valType: INT
            }
        """


class Int64VectorFeatureType(FeatureType):
    def to_feature_config(self) -> str:
        return """
           type: {
                type: TENSOR
                tensorCategory: DENSE
                dimensionType: [INT]
                valType: LONG
            }
        """


class DoubleVectorFeatureType(FeatureType):
    def to_feature_config(self) -> str:
        return """
           type: {
                type: TENSOR
                tensorCategory: DENSE
                dimensionType: [INT]
                valType: DOUBLE
            }
        """
# tensor dimension/axis
class Dimension:
    def __init__(self, shape: int, dType: ValueType = ValueType.INT32):
        self.shape = shape
        self.dType = dType


BOOLEAN = BooleanFeatureType()
INT32 = Int32FeatureType()
INT64 = Int64FeatureType()
FLOAT = FloatFeatureType()
DOUBLE = DoubleFeatureType()
STRING = StringFeatureType()
BYTES = BytesFeatureType()
FLOAT_VECTOR = FloatVectorFeatureType()
INT32_VECTOR = Int32VectorFeatureType()
INT64_VECTOR = Int64VectorFeatureType()
DOUBLE_VECTOR = DoubleVectorFeatureType()


from typing import List, Optional, Union

from jinja2 import Template

from feathr.typed_key import TypedKey


class FeatureQuery:
    """A FeatureQuery contains a list of features
    Attributes:
        feature_list: a list of feature names
        key: key of `feature_list`, all features must share the same key
        """
    def __init__(self, feature_list: List[str], key: Optional[Union[TypedKey, List[TypedKey]]] = None) -> None:
        self.key = key
        if isinstance(key, TypedKey):
            self.key = [key]
        self.feature_list = feature_list

    def to_config(self) -> str:
        tm = Template("""
            {
                key: {{key_columns}}
                featureList: [{{feature_names}}]
            }
        """)
        key_columns = ", ".join(k.key_column for k in self.key) if self.key else ["NOT_NEEDED"]
        feature_list = ", ".join(f for f in self.feature_list)
        return tm.render(key_columns = key_columns, feature_names = feature_list)



from abc import ABC, abstractmethod
from typing import List, Optional
from jinja2 import Template


class Sink(ABC):
    @abstractmethod
    def to_write_config(self) -> str:
        pass


class RedisSink(Sink):
    def __init__(self, table_name: str) -> None:
        self.table_name = table_name

    def to_write_config(self) -> str:
        """Produce the config used in feature materialization"""
        tm = Template("""  
            {
                name: REDIS
                params: {
                    table_name: "{{source.table_name}}"
                }
            }
        """)
        msg = tm.render(source=self)
        return msg


import os
from datetime import datetime, timedelta
from pathlib import Path

from click.testing import CliRunner
from feathr import BOOLEAN, FLOAT, INT32, ValueType
from feathr.client import FeathrClient
from feathr import ValueType
from feathr.job_utils import get_result_df
from feathr import (BackfillTime, MaterializationSettings)
from feathr import FeatureQuery
from feathr import ObservationSettings
from feathr import RedisSink
from feathr import TypedKey
from feathrcli.cli import init
import pytest

from test_fixture import basic_test_setup
# make sure you have run the upload feature script before running these tests
# the feature configs are from feathr_project/data/feathr_user_workspace
def test_feathr_online_store_agg_features():
    """
    Test FeathrClient() get_online_features and batch_get can get data correctly.
    """

    # use different time for testing to avoid write conflicts
    now = datetime.now()

    online_test_table = ''.join(['nycTaxiCITable','_', str(now.minute), '_', str(now.second)])
    test_workspace_dir = Path(
        __file__).parent.resolve() / "test_user_workspace"
    # os.chdir(test_workspace_dir)

    client = basic_test_setup(os.path.join(test_workspace_dir, "feathr_config.yaml"))

    backfill_time = BackfillTime(start=datetime(
        2020, 5, 20), end=datetime(2020, 5, 20), step=timedelta(days=1))
    redisSink = RedisSink(table_name=online_test_table)
    settings = MaterializationSettings("nycTaxiTable",
                                       sinks=[redisSink],
                                       feature_names=[
                                           "f_location_avg_fare", "f_location_max_fare"],
                                       backfill_time=backfill_time)
    client.materialize_features(settings)
    # just assume the job is successful without validating the actual result in Redis. Might need to consolidate
    # this part with the test_feathr_online_store test case
    client.wait_job_to_finish(timeout_sec=600)

    res = client.get_online_features(online_test_table, '265', [
                                     'f_location_avg_fare', 'f_location_max_fare'])
    # just assme there are values. We don't hard code the values for now for testing
    # the correctness of the feature generation should be garunteed by feathr runtime.
    # ID 239 and 265 are available in the `DOLocationID` column in this file:
    # https://s3.amazonaws.com/nyc-tlc/trip+data/green_tripdata_2020-04.csv
    # View more detials on this dataset: https://www1.nyc.gov/site/tlc/about/tlc-trip-record-data.page
    assert len(res) == 2
    assert res[0] != None
    assert res[1] != None
    res = client.multi_get_online_features(online_test_table,
                                           ['239', '265'],
                                           ['f_location_avg_fare', 'f_location_max_fare'])
    assert res['239'][0] != None
    assert res['239'][1] != None
    assert res['265'][0] != None
    assert res['265'][1] != None

@pytest.mark.skip(reason="Add back when complex types are supported in python API")
def test_feathr_online_store_non_agg_features():
    """
    Test FeathrClient() online_get_features and batch_get can get data correctly.
    """
    test_workspace_dir = Path(
        __file__).parent.resolve() / "test_user_workspace"
    client = basic_test_setup(os.path.join(test_workspace_dir, "feathr_config.yaml"))
    now = datetime.now()
    online_test_table = ''.join(['nycTaxiCITable','_', str(now.minute), '_', str(now.second)])
    backfill_time = BackfillTime(start=datetime(
        2020, 5, 20), end=datetime(2020, 5, 20), step=timedelta(days=1))
    redisSink = RedisSink(table_name=online_test_table)
    settings = MaterializationSettings("nycTaxiTable",
                                       sinks=[redisSink],
                                       feature_names=["f_gen_trip_distance", "f_gen_is_long_trip_distance", "f1", "f2", "f3", "f4", "f5", "f6"],
                                       backfill_time=backfill_time)

    client.materialize_features(settings)
    # just assume the job is successful without validating the actual result in Redis. Might need to consolidate
    # this part with the test_feathr_online_store test case
    client.wait_job_to_finish(timeout_sec=600)

    res = client.get_online_features(online_test_table, '111', ['f_gen_trip_distance', 'f_gen_is_long_trip_distance',
                                                                   'f1', 'f2', 'f3', 'f4', 'f5', 'f6'])
    # just assme there are values. We don't hard code the values for now for testing
    # the correctness of the feature generation should be garunteed by feathr runtime.
    # ID 239 and 265 are available in the `DOLocationID` column in this file:
    # https://s3.amazonaws.com/nyc-tlc/trip+data/green_tripdata_2020-04.csv
    # View more detials on this dataset: https://www1.nyc.gov/site/tlc/about/tlc-trip-record-data.page

    assert len(res) == 8
    assert res[0] != None
    assert res[1] != None
    # assert constant features
    _validate_constant_feature(res)
    res = client.multi_get_online_features(online_test_table,
                                           ['239', '265'],
                                           ['f_gen_trip_distance', 'f_gen_is_long_trip_distance', 'f1', 'f2', 'f3', 'f4', 'f5', 'f6'])
    _validate_constant_feature(res['239'])
    assert res['239'][0] != None
    assert res['239'][1] != None
    _validate_constant_feature(res['265'])
    assert res['265'][0] != None
    assert res['265'][1] != None


def _validate_constant_feature(feature):
    assert feature[2] == [10.0, 20.0, 30.0]
    assert feature[3] == ['a', 'b', 'c']
    assert feature[4] == ([1, 2, 3], ['10', '20', '30'])
    assert feature[5] == ([1, 2, 3], [True, False, True])
    assert feature[6] == ([1, 2, 3], [1.0, 2.0, 3.0])
    assert feature[7] == ([1, 2, 3], [1, 2, 3])


def test_feathr_get_offline_features():
    """
    Test get_offline_features() can get data correctly.
    """
    runner = CliRunner()
    with runner.isolated_filesystem():
        runner.invoke(init, [])
        client = basic_test_setup(
            "./feathr_user_workspace/feathr_config.yaml")

        location_id = TypedKey(key_column="DOLocationID",
                               key_column_type=ValueType.INT32,
                               description="location id in NYC",
                               full_name="nyc_taxi.location_id")

        feature_query = FeatureQuery(
            feature_list=["f_location_avg_fare"], key=location_id)
        settings = ObservationSettings(
            observation_path="abfss://feathrazuretest3fs@feathrazuretest3storage.dfs.core.windows.net/demo_data/green_tripdata_2020-04.csv",
            event_timestamp_column="lpep_dropoff_datetime",
            timestamp_format="yyyy-MM-dd HH:mm:ss")

        now = datetime.now()
        # set output folder based on different runtime
        if client.spark_runtime == 'databricks':
            output_path = ''.join(['dbfs:/feathrazure_cijob','_', str(now.minute), '_', str(now.second), ".avro"])
        else:
            output_path = ''.join(['abfss://feathrazuretest3fs@feathrazuretest3storage.dfs.core.windows.net/demo_data/output','_', str(now.minute), '_', str(now.second), ".avro"])

        
        client.get_offline_features(observation_settings=settings,
                                    feature_query=feature_query,
                                    output_path=output_path)

        # assuming the job can successfully run; otherwise it will throw exception
        client.wait_job_to_finish(timeout_sec=900)
        
        # download result and just assert the returned result is not empty
        res_df = get_result_df(client)
        assert res_df.shape[0] > 0



from feathrcli.cli import init
from click.testing import CliRunner
from feathr.client import FeathrClient
import os
import glob
from test_fixture import basic_test_setup


def test_feathr_register_features_e2e():
    runner = CliRunner()
    with runner.isolated_filesystem():
        runner.invoke(init, [])
        client = basic_test_setup(
            "./feathr_user_workspace/feathr_config.yaml")
        client.register_features()
        # in CI test, the project name is set by the CI pipeline so we don't know it here. Just get all the features to make sure it works
        all_features = client.list_registered_features()
        assert 'f_is_long_trip_distance' in all_features


def test_feathr_get_features_from_registry():
    """
    Test FeathrClient() sync features and get all the conf files from registry
    """
    runner = CliRunner()
    with runner.isolated_filesystem():
        result = runner.invoke(init, [])

        assert result.exit_code == 0
        assert os.path.isdir("./feathr_user_workspace")
        os.chdir('feathr_user_workspace')

        # Look for conf files, we shouldn't have any
        total_conf_files = glob.glob('*/*.conf', recursive=True)
        assert len(total_conf_files) == 0
        
        client = FeathrClient()
        # Sync workspace from registry, will get all conf files back
        client.get_features_from_registry("frame_getting_started")

        total_conf_files = glob.glob('*/*.conf', recursive=True)
        # we should have at least 3 conf files
        assert len(total_conf_files) == 3



from typing import List, Optional, Union

from jinja2 import Template

from feathr.dtype import FeatureType
from feathr.feature import FeatureBase
from feathr.transformation import RowTransformation
from feathr.typed_key import DUMMY_KEY, TypedKey


class DerivedFeature(FeatureBase):
    """A derived feature is a feature defined on top of other features, rather than external data source.
    Attributes:
        name: derived feature name
        feature_type: type of derived feature
        key: All features with corresponding keys that this derived feature depends on
        input_features: features that this derived features depends on
        transform: transformation that produces the derived feature value, based on the input_features
    """

    def __init__(self,
                name: str,
                feature_type: FeatureType,
                input_features: Union[FeatureBase, List[FeatureBase]],
                transform: Union[str, RowTransformation],
                key: Optional[Union[TypedKey, List[TypedKey]]] = [DUMMY_KEY]):
        super(DerivedFeature, self).__init__(name, feature_type, key=key, transform=transform)
        self.input_features = input_features if isinstance(input_features, List) else [input_features]
        self.validate_feature()

    def validate_feature(self):
        """Validate the derived feature is valid"""
        # Validate key alias
        input_feature_key_alias = []
        for feature in self.input_features:
            input_feature_key_alias.extend(feature.key_alias)
        for key_alias in self.key_alias:
            assert key_alias in input_feature_key_alias, "key alias {} in derived feature {} must come from " \
                   "its input features key alias list {}".format(key_alias, self.name, input_feature_key_alias)


    def to_feature_config(self) -> str:
        tm = Template("""
            {{derived_feature.name}}: {
                key: [{{','.join(derived_feature.key_alias)}}]
                inputs: {
                    {% for feature in derived_feature.input_features %}
                        {{feature.feature_alias}}: {
                            key: [{{','.join(feature.key_alias)}}],
                            feature: {{feature.name}}
                        }
                    {% endfor %}
                }
                definition: {{derived_feature.transform.to_feature_config()}}
                {{derived_feature.feature_type.to_feature_config()}}
            }
        """)
        return tm.render(derived_feature=self)



import glob
import importlib
import os
import sys
from pathlib import Path
from typing import List, Optional

from jinja2 import Template
from loguru import logger
from pyapacheatlas.auth import ServicePrincipalAuthentication
from pyapacheatlas.core import (AtlasEntity, AtlasProcess, PurviewClient,
                                TypeCategory)
from pyapacheatlas.core.typedef import (AtlasAttributeDef, EntityTypeDef,
                                        RelationshipTypeDef)
from pyapacheatlas.core.util import GuidTracker
from pyexpat import features
from pyhocon import ConfigFactory

from feathr._envvariableutil import _EnvVaraibleUtil
from feathr._file_utils import write_to_file
from feathr.anchor import FeatureAnchor
from feathr.feature import Feature
from feathr.feature_derivations import DerivedFeature
from feathr.repo_definitions import RepoDefinitions
from feathr.source import Source
from feathr.transformation import Transformation


class _FeatureRegistry():

    def __init__(self, config_path):
        """
        Initializes the feature registry, doing the following:
        - Use an Azure Service Principal to communicate with Azure Purview
        - Initialize an Azure Purview Client
        - Initialize the GUID tracker, project name, etc.
        """
        envutils = _EnvVaraibleUtil(config_path)
        self.project_name = envutils.get_environment_variable_with_default('project_config', 'project_name')
        self.FEATURE_REGISTRY_DELIMITER = envutils.get_environment_variable_with_default('feature_registry', 'purview', 'delimiter')
        self.azure_purview_name = envutils.get_environment_variable_with_default('feature_registry', 'purview', 'purview_name')

        self.oauth = ServicePrincipalAuthentication(
            tenant_id=_EnvVaraibleUtil.get_environment_variable(
                "AZURE_TENANT_ID"),
            client_id=_EnvVaraibleUtil.get_environment_variable(
                "AZURE_CLIENT_ID"),
            client_secret=_EnvVaraibleUtil.get_environment_variable(
                "AZURE_CLIENT_SECRET")
        )
        self.purview_client = PurviewClient(
            account_name=self.azure_purview_name,
            authentication=self.oauth
        )
        self.guid = GuidTracker(starting=-1000)
        self.entity_batch_queue = []

        

    def _register_feathr_feature_types(self):
        """
        Register the feathr types if we haven't done so. Note that this only needs to be called once per provisioning
        a system. Basically this function registers all the feature type definition in a Atlas compatible system.
        """

        # COMMAND ----------

        # Set up a few types and relationships
        # This is a one time thing but necessary to make the demo work
        # It also demonstrates how you can capture different attributes
        # for your dataframes, dataframe columns, and jobs.

        type_feathr_features = EntityTypeDef(
            name="feathr_workspace",
            attributeDefs=[
                AtlasAttributeDef(
                    name="sources", typeName="array<feathr_source>", cardinality="SET"),
                AtlasAttributeDef(
                    name="anchors", typeName="array<feathr_anchor>", cardinality="SET"),
                AtlasAttributeDef(
                    name="derivations", typeName="array<feathr_derivation>", cardinality="SET"),
                AtlasAttributeDef(name="raw_hocon_feature_definition_config",
                                  typeName="string", cardinality="SINGLE"),
                AtlasAttributeDef(name="raw_hocon_feature_join_config",
                                  typeName="string", cardinality="SINGLE"),
                AtlasAttributeDef(name="raw_hocon_feature_generation_config",
                                  typeName="string", cardinality="SINGLE"),

            ],
            superTypes=["DataSet"],
            serviceType="Feathr Workspace"
        )
        type_feathr_sources = EntityTypeDef(
            name="feathr_source",
            attributeDefs=[
                AtlasAttributeDef(name="isTimeSeries",
                                  typeName="boolean", cardinality="SINGLE"),
                AtlasAttributeDef(name="type", typeName="string",
                                  cardinality="SINGLE"),
                AtlasAttributeDef(
                    name="location", typeName="string", cardinality="SINGLE"),
                AtlasAttributeDef(name="timestamp",
                                  typeName="string", cardinality="SINGLE"),
                AtlasAttributeDef(name="timestamp_format",
                                  typeName="string", cardinality="SINGLE"),
                AtlasAttributeDef(name="timestampColumn",
                                  typeName="string", cardinality="SINGLE"),
                AtlasAttributeDef(name="timestampColumnFormat",
                                  typeName="string", cardinality="SINGLE"),
            ],
            superTypes=["DataSet"],
            serviceType="Feathr Feature Source"
        )

        type_feathr_derivations = EntityTypeDef(
            name="feathr_derivation",
            attributeDefs=[
                AtlasAttributeDef(name="key", typeName="string",
                                  cardinality="SINGLE"),
                AtlasAttributeDef(name="inputs", typeName="string",
                                  cardinality="SINGLE"),
                AtlasAttributeDef(name="definition",
                                  typeName="string", cardinality="SINGLE"),
                AtlasAttributeDef(name="type", typeName="string",
                                  cardinality="SINGLE"),
                AtlasAttributeDef(name="tensorCategory",
                                  typeName="string", cardinality="SINGLE"),
                AtlasAttributeDef(name="dimensionType",
                                  typeName="string", cardinality="SINGLE"),
                AtlasAttributeDef(
                    name="valType", typeName="string", cardinality="SINGLE"),
            ],
            superTypes=["DataSet"],
            serviceType="Feathr Feature Derivation"
        )

        type_feathr_anchors = EntityTypeDef(
            name="feathr_anchor",
            attributeDefs=[
                AtlasAttributeDef(
                    name="source", typeName="string", cardinality="SINGLE"),
                AtlasAttributeDef(name="extractor",
                                  typeName="string", cardinality="SINGLE"),
                AtlasAttributeDef(
                    name="features", typeName="array<feathr_anchor_feature>", cardinality="SET"),
                AtlasAttributeDef(name="key", typeName="string",
                                  cardinality="SINGLE"),
                AtlasAttributeDef(name="key.sqlExpr",
                                  typeName="string", cardinality="SINGLE"),
                AtlasAttributeDef(name="timestampColumn",
                                  typeName="string", cardinality="SINGLE"),
                AtlasAttributeDef(name="timestampColumnFormat",
                                  typeName="string", cardinality="SINGLE"),
                AtlasAttributeDef(
                    name="location", typeName="string", cardinality="SINGLE"),
            ],
            superTypes=["DataSet"],
            serviceType="Feathr Anchors"
        )

        type_feathr_anchor_features = EntityTypeDef(
            name="feathr_anchor_feature",
            attributeDefs=[
                AtlasAttributeDef(name="def", typeName="string",
                                  cardinality="SINGLE"),
                AtlasAttributeDef(name="aggregation",
                                  typeName="string", cardinality="SINGLE"),
                AtlasAttributeDef(name="window", typeName="string",
                                  cardinality="SINGLE"),
                AtlasAttributeDef(
                    name="groupBy", typeName="string", cardinality="SINGLE"),
                AtlasAttributeDef(
                    name="default", typeName="string", cardinality="SINGLE"),
                AtlasAttributeDef(name="def.sqlExpr",
                                  typeName="string", cardinality="SINGLE"),
                AtlasAttributeDef(name="type", typeName="string",
                                  cardinality="SINGLE"),
                AtlasAttributeDef(name="tensorCategory",
                                  typeName="string", cardinality="SINGLE"),
                AtlasAttributeDef(name="dimensionType",
                                  typeName="string", cardinality="SINGLE"),
                AtlasAttributeDef(
                    name="valType", typeName="string", cardinality="SINGLE"),
            ],
            superTypes=["DataSet"],
            serviceType="Feathr Feature"
        )

        feathr_source_to_feature_relationship = RelationshipTypeDef(
            name="feathr_source_to_feature_relationship",
            relationshipCategory="COMPOSITION",
            endDef1={
                "type": "feathr_workspace",
                "name": "feathr_feature",
                "isContainer": True,
                "cardinality": "SET",
                "isLegacyAttribute": False
            },
            endDef2={
                "type": "feathr_source",
                "name": "feathr_source",
                "isContainer": False,
                "cardinality": "SINGLE",
                "isLegacyAttribute": False
            }
        )

        feathr_anchor_to_feature_relationship = RelationshipTypeDef(
            name="feathr_anchor_to_feature_relationship",
            relationshipCategory="COMPOSITION",
            endDef1={
                "type": "feathr_workspace",
                "name": "feathr_feature",
                "isContainer": True,
                "cardinality": "SET",
                "isLegacyAttribute": False
            },
            endDef2={
                "type": "feathr_anchor",
                "name": "feathr_anchor",
                "isContainer": False,
                "cardinality": "SINGLE",
                "isLegacyAttribute": False
            }
        )

        feathr_derivation_to_feature_relationship = RelationshipTypeDef(
            name="feathr_derivation_to_feature_relationship",
            relationshipCategory="COMPOSITION",
            endDef1={
                "type": "feathr_workspace",
                "name": "feathr_feature",
                "isContainer": True,
                "cardinality": "SET",
                "isLegacyAttribute": False
            },
            endDef2={
                "type": "feathr_derivation",
                "name": "feathr_derivation",
                "isContainer": False,
                "cardinality": "SINGLE",
                "isLegacyAttribute": False
            }
        )

        typedef_results = self.purview_client.upload_typedefs(
            entityDefs=[type_feathr_anchor_features, type_feathr_anchors,
                        type_feathr_derivations, type_feathr_sources, type_feathr_features],
            relationshipDefs=[feathr_source_to_feature_relationship,
                              feathr_anchor_to_feature_relationship, feathr_derivation_to_feature_relationship],
            force_update=True)
        logger.info("Feature Type System Initialized.")

    def _parse_anchor_features(self, anchor_name: str) -> List[AtlasEntity]:
        """
        This function will parse the actual features inside an anchor

        Args:
            anchor_name (str): name of the anchor

        Returns:
            list[AtlasEntity]: a list of the parsed actual features inside an anchor definition
        """
        anchor_feature_batch = []
        for anchor_feature_name in self.feathr_feature_config["anchors"][anchor_name]["features"]:
            # there will be cases where we don't have a "def" field in this hocon file. First we detect whether it
            # has a "def" field. If the anchor feature does have this field, we will use this field directly otherwise
            res = self.feathr_feature_config["anchors"][anchor_name]["features"][anchor_feature_name]
            if isinstance(res, str):
                anchor_feature = AtlasEntity(
                    name=anchor_feature_name,
                    qualified_name=self.project_name + self.FEATURE_REGISTRY_DELIMITER +
                    anchor_name + self.FEATURE_REGISTRY_DELIMITER + anchor_feature_name,
                    attributes={
                        "def": res,
                    },
                    typeName="feathr_anchor_feature",
                    guid=self.guid.get_guid(),
                )
            else:
                feature_defstr = self.feathr_feature_config["anchors"][
                    anchor_name]["features"][anchor_feature_name]
                anchor_feature = AtlasEntity(
                    name=anchor_feature_name,
                    qualified_name=self.project_name + self.FEATURE_REGISTRY_DELIMITER +
                    anchor_name + self.FEATURE_REGISTRY_DELIMITER + anchor_feature_name,
                    attributes={
                        "def": str(feature_defstr),
                        "aggregation": self.feathr_feature_config["anchors"][anchor_name]["features"][anchor_feature_name].get_string("aggregation", ""),
                        "window": self.feathr_feature_config["anchors"][anchor_name]["features"][anchor_feature_name].get_string("window", ""),
                        "groupBy": self.feathr_feature_config["anchors"][anchor_name]["features"][anchor_feature_name].get_string("groupBy", ""),
                        "default": self.feathr_feature_config["anchors"][anchor_name]["features"][anchor_feature_name].get_string("default", ""),
                        "def.sqlExpr": self.feathr_feature_config["anchors"][anchor_name]["features"][anchor_feature_name].get_string("def.sqlExpr", ""),
                        "type": self.feathr_feature_config["anchors"][anchor_name]["features"][anchor_feature_name].get_string("type", ""),
                        "tensorCategory": self.feathr_feature_config["anchors"][anchor_name]["features"][anchor_feature_name].get_string("tensorCategory", ""),
                        "dimensionType": self.feathr_feature_config["anchors"][anchor_name]["features"][anchor_feature_name].get_string("dimensionType", ""),
                        "valType": self.feathr_feature_config["anchors"][anchor_name]["features"][anchor_feature_name].get_string("valType", ""),
                    },
                    typeName="feathr_anchor_feature",
                    guid=self.guid.get_guid(),
                )

            self.entity_batch_queue.append(anchor_feature)
            anchor_feature_batch.append(anchor_feature)
        return anchor_feature_batch

    def _parse_anchors(self, feathr_anchors: str) -> List[AtlasEntity]:
        """
        parse the actual features inside an anchor

        Args:
            feathr_anchors (str): name of the feathr anchor

        Returns:
            list[AtlasEntity]: a list of the parsed anchors. Note that this is not the actual anchor feature.
        """

        anchors_batch = []

        # note that we want to put anchor_feature_batch in the same level as anchors_batch
        # because we want to make sure all the anchor features are stored and captured globally

        for i, anchor_name in enumerate(feathr_anchors):
            # First, parse all the features in this anchor
            anchor_feature_batch = self._parse_anchor_features(anchor_name)

            # Second, get all the other attributes from the anchors
            anchor = AtlasEntity(
                name=anchor_name,
                qualified_name=self.project_name + self.FEATURE_REGISTRY_DELIMITER + anchor_name,
                attributes={
                    "source": str(self.feathr_feature_config["anchors"][anchor_name]["source"]),
                    "key": str(self.feathr_feature_config["anchors"][anchor_name]["key"]),
                    "features": [s.to_json(minimum=True) for s in anchor_feature_batch],
                    "extractor": str(self.feathr_feature_config["anchors"][anchor_name].get_string("extractor", "")),
                    "key.sqlExpr": str(self.feathr_feature_config["anchors"][anchor_name].get_string("key.sqlExpr", "")),
                    "timestampColumn": str(self.feathr_feature_config["anchors"][anchor_name].get_string("timestampColumn", "")),
                    "timestampColumnFormat": str(self.feathr_feature_config["anchors"][anchor_name].get_string("timestampColumnFormat", "")),
                    "location": str(self.feathr_feature_config["anchors"][anchor_name].get_string("location", "")),
                },
                typeName="feathr_anchor",
                guid=self.guid.get_guid(),
            )

            # add Atlas Process for feature lineage
            for individual_anchor_feature in anchor_feature_batch:
                lineage_process = AtlasProcess(
                    name=individual_anchor_feature.name + " to " + anchor_name,
                    typeName="Process",
                    qualified_name=self.FEATURE_REGISTRY_DELIMITER + "PROCESS" + self.FEATURE_REGISTRY_DELIMITER + self.project_name +
                    self.FEATURE_REGISTRY_DELIMITER + anchor_name + self.FEATURE_REGISTRY_DELIMITER +
                    individual_anchor_feature.name,
                    inputs=[individual_anchor_feature],
                    outputs=[anchor],
                    guid=self.guid.get_guid(),
                )
                self.entity_batch_queue.append(lineage_process)

            self.entity_batch_queue.append(anchor)
            anchors_batch.append(anchor)
        return anchors_batch

    def _parse_sources(self, feathr_sources: str) -> List[AtlasEntity]:
        """
        parse the `sources` section of the feature configuration

        Args:
            feathr_sources (str): the name of the source

        Returns:
            list[AtlasEntity]: list of the parsed sources
        """
        sources_batch = []
        for i, source_name in enumerate(feathr_sources):

            source = AtlasEntity(
                name=source_name,
                qualified_name=self.project_name + self.FEATURE_REGISTRY_DELIMITER + source_name,
                attributes={
                    "location": str(self.feathr_feature_config["sources"][source_name]["location"]["path"]),
                    "type": str(self.feathr_feature_config["sources"][source_name].get_string("type", "")),
                    "isTimeSeries": bool(self.feathr_feature_config["sources"][source_name].get_string("isTimeSeries", "")),
                    "timestamp": str(self.feathr_feature_config["sources"][source_name].get_string("timeWindowParameters.timestamp", "")),
                    "timestamp_format": str(self.feathr_feature_config["sources"][source_name].get_string("timeWindowParameters.timestamp_format", "")),
                    "timestampColumn": str(self.feathr_feature_config["sources"][source_name].get_string("timeWindowParameters.timestampColumn", "")),
                    "timestampColumnFormat": str(self.feathr_feature_config["sources"][source_name].get_string("timeWindowParameters.timestampColumnFormat", ""))
                },
                typeName="feathr_source",
                guid=self.guid.get_guid(),
            )
            self.entity_batch_queue.append(source)
            sources_batch.append(source)
        return sources_batch

    def _parse_derivations(self, feathr_derivations):
        derivations_batch = []

        for i, derivation_name in enumerate(feathr_derivations):

            derivation = AtlasEntity(
                name=derivation_name,
                qualified_name=self.project_name +
                self.FEATURE_REGISTRY_DELIMITER + derivation_name,
                attributes={
                    "key": str(self.feathr_feature_config["derivations"][derivation_name].get_string("key", "")),
                    "type": str(self.feathr_feature_config["derivations"][derivation_name].get_string("type", "")),
                    "inputs": bool(self.feathr_feature_config["derivations"][derivation_name].get_string("inputs", "")),
                    "definition": str(self.feathr_feature_config["derivations"][derivation_name].get_string("definition", "")),
                    "tensorCategory": str(self.feathr_feature_config["derivations"][derivation_name].get_string("tensorCategory", "")),
                    "dimensionType": str(self.feathr_feature_config["derivations"][derivation_name].get_string("dimensionType", "")),
                    "valType": str(self.feathr_feature_config["derivations"][derivation_name].get_string("valType", ""))
                },
                typeName="feathr_derivation",
                guid=self.guid.get_guid(),
            )
            self.entity_batch_queue.append(derivation)
            derivations_batch.append(derivation)
        return derivations_batch

    def _read_config_from_workspace(self, workspace_path: str):
        """
        Read a HOCON file from a workspace
        """
        raw_hocon_feature_definition_config = ""
        raw_hocon_feature_join_config = ""
        raw_hocon_feature_generation_config = ""

        # get feature configuration file
        feature_config_paths = glob.glob(
            os.path.join(workspace_path, "feature_conf", '*.conf'))
        logger.info("Reading feature configuration from {}",
                    feature_config_paths)
        if len(feature_config_paths) > 0:
            feature_config_path = feature_config_paths[0]
            self.feathr_feature_config = ConfigFactory.parse_file(
                feature_config_path)
            with open(feature_config_path, "r") as f:
                raw_hocon_feature_definition_config = f.read()

            # get feature join config file
            feature_join_paths = glob.glob(os.path.join(
                workspace_path, "feature_join_conf", '*.conf'))
            logger.info("Reading feature join configuration from {}",
                    feature_join_paths)
        if len(feature_join_paths) > 0:
            feature_join_path = feature_join_paths[0]
            self.feathr_feature_join = ConfigFactory.parse_file(feature_join_path)
            with open(feature_join_path, "r") as f:
                raw_hocon_feature_join_config = f.read()

        # get feature generation config file
        feature_generation_paths = glob.glob(
            os.path.join(workspace_path, "feature_conf", '*.conf'))
        logger.info("Reading feature generation configuration from {}",
                    feature_generation_paths)
        if len(feature_generation_paths) > 0:
            feature_generation_path = feature_generation_paths[0]
            self.feathr_feature_generation = ConfigFactory.parse_file(
                feature_generation_path)
            with open(feature_generation_path, "r") as f:
                raw_hocon_feature_generation_config = f.read()

        feathr_anchors = self.feathr_feature_config.get("anchors", "")
        feathr_sources = self.feathr_feature_config.get("sources", "")
        feathr_derivations = self.feathr_feature_config.get("derivations", "")

        sources_batch = []
        anchors_batch = []
        # parse all the anchors
        if feathr_anchors:
            anchors_batch = self._parse_anchors(feathr_anchors)

        if feathr_sources:
            sources_batch = self._parse_sources(feathr_sources)

        if feathr_derivations:
            derivations_batch = self._parse_derivations(feathr_derivations)
            workspace = AtlasEntity(
                name=self.project_name,
                qualified_name=self.project_name,
                attributes={
                    "sources": [s.to_json(minimum=True) for s in sources_batch],
                    "anchors": [s.to_json(minimum=True) for s in anchors_batch],
                    "derivations": [s.to_json(minimum=True) for s in derivations_batch],
                    "raw_hocon_feature_definition_config": raw_hocon_feature_definition_config,
                    "raw_hocon_feature_join_config": raw_hocon_feature_join_config,
                    "raw_hocon_feature_generation_config": raw_hocon_feature_generation_config,
                },
                typeName="feathr_workspace",
                guid=self.guid.get_guid(),
            )
        else:
            derivations_batch = self._parse_derivations(feathr_derivations)
            workspace = AtlasEntity(
                name=self.project_name,
                qualified_name=self.project_name,
                attributes={
                    "sources": [s.to_json(minimum=True) for s in sources_batch],
                    "anchors": [s.to_json(minimum=True) for s in anchors_batch],
                    "raw_hocon_feature_definition_config": raw_hocon_feature_definition_config,
                    "raw_hocon_feature_join_config": raw_hocon_feature_join_config,
                    "raw_hocon_feature_generation_config": raw_hocon_feature_generation_config,
                },
                typeName="feathr_workspace",
                guid=self.guid.get_guid(),
            )

        # add lineage from anchor to workspace
        for individual_anchor in anchors_batch:
            lineage_process = AtlasProcess(
                name=individual_anchor.name + " to " + self.project_name,
                typeName="Process",
                qualified_name=self.FEATURE_REGISTRY_DELIMITER + "PROCESS" + self.FEATURE_REGISTRY_DELIMITER + self.project_name +
                self.FEATURE_REGISTRY_DELIMITER + individual_anchor.name,
                inputs=[individual_anchor],
                outputs=[workspace],
                guid=self.guid.get_guid(),
            )
            self.entity_batch_queue.append(lineage_process)

        # add lineage from source to workspace
        for individual_source in sources_batch:
            lineage_process = AtlasProcess(
                name=individual_source.name + " to " + self.project_name,
                typeName="Process",
                qualified_name=self.FEATURE_REGISTRY_DELIMITER + "PROCESS" + self.FEATURE_REGISTRY_DELIMITER + self.project_name +
                self.FEATURE_REGISTRY_DELIMITER + individual_source.name,
                inputs=[individual_source],
                outputs=[workspace],
                guid=self.guid.get_guid(),
            )
            self.entity_batch_queue.append(lineage_process)

        if feathr_derivations:
            # add lineage from derivation to workspace
            for individual_derivation in derivations_batch:
                lineage_process = AtlasProcess(
                    name=individual_derivation.name + " to " + self.project_name,
                    typeName="Process",
                    qualified_name=self.FEATURE_REGISTRY_DELIMITER + "PROCESS" + self.FEATURE_REGISTRY_DELIMITER + self.project_name +
                    self.FEATURE_REGISTRY_DELIMITER + individual_derivation.name,
                    inputs=[individual_derivation],
                    outputs=[workspace],
                    guid=self.guid.get_guid(),
                )
                self.entity_batch_queue.append(lineage_process)

        self.entity_batch_queue.append(workspace)

    @classmethod
    def _get_py_files(self, path: Path) -> List[Path]:
        """Get all Python files under path recursively, excluding __init__.py"""
        py_files = []
        for item in path.glob('**/*.py'):
            if "__init__.py" != item.name:
                py_files.append(item)
        return py_files

    @classmethod
    def _convert_to_module_path(self, path: Path, workspace_path: Path) -> str:
        """Convert a Python file path to its module path so that we can import it later"""
        prefix = os.path.commonprefix([path.resolve(), workspace_path.resolve()])
        resolved_path = str(path.resolve())
        module_path = resolved_path[len(prefix): -len(".py")]
        # Convert features under nested folder to module name
        # e.g. /path/to/pyfile will become path.to.pyfile
        return (
            module_path
            .lstrip('/')
            .replace("/", ".")
        )

    @classmethod
    def _extract_features_from_context(self, anchor_list, derived_feature_list, result_path: Path) -> RepoDefinitions:
        """Collect feature definitions from the context instead of python files"""
        definitions = RepoDefinitions(
            sources=set(),
            features=set(),
            transformations=set(),
            feature_anchors=set(),
            derived_features=set()
        )
        for derived_feature in derived_feature_list:
            if isinstance(derived_feature, DerivedFeature):                    
                definitions.derived_features.add(derived_feature)
                definitions.transformations.add(vars(derived_feature)["transform"])
            else:
                RuntimeError("Object cannot be parsed. `derived_feature_list` should be a list of `DerivedFeature`.")

        for anchor in anchor_list:
            # obj is `FeatureAnchor`
            definitions.feature_anchors.add(anchor)
            # add the source section of this `FeatureAnchor` object
            definitions.sources.add(vars(anchor)['source'])
            for feature in vars(anchor)['features']:
                # get the transformation object from `Feature` or `DerivedFeature`
                if isinstance(anchor, Feature):
                    # feature is of type `Feature` 
                    definitions.features.add(anchor)
                    definitions.transformations.add(vars(feature)["transform"])
                else:
                    RuntimeError("Object cannot be parsed.")
        
        return definitions


    @classmethod
    def _extract_features(self, workspace_path: Path) -> RepoDefinitions:
        """Collect feature definitions from the python file, convert them into feature config and save them locally"""
        os.chdir(workspace_path)
        # Add workspace path to system path so that we can load features defined in Python via import_module
        sys.path.append(str(workspace_path))
        definitions = RepoDefinitions(
            sources=set(),
            features=set(),
            transformations=set(),
            feature_anchors=set(),
            derived_features=set()
        )
        for py_file in self._get_py_files(workspace_path):
            module_path = self._convert_to_module_path(py_file, workspace_path)
            module = importlib.import_module(module_path)
            for attr_name in dir(module):
                obj = getattr(module, attr_name)
                if isinstance(obj, Source):
                    definitions.sources.add(obj)
                elif isinstance(obj, Feature):
                    definitions.features.add(obj)
                elif isinstance(obj, DerivedFeature):
                    definitions.derived_features.add(obj)
                elif isinstance(obj, FeatureAnchor):
                    definitions.feature_anchors.add(obj)
                elif isinstance(obj, Transformation):
                    definitions.transformations.add(obj)
        return definitions

    @classmethod
    def save_to_feature_config(self, workspace_path: Path):
        """Save feature definition within the workspace into HOCON feature config files"""
        repo_definitions = self._extract_features(workspace_path)
        self._save_request_feature_config(repo_definitions)
        self._save_anchored_feature_config(repo_definitions)
        self._save_derived_feature_config(repo_definitions)

    @classmethod
    def save_to_feature_config_from_context(self, anchor_list, derived_feature_list, local_workspace_dir: Path):
        """Save feature definition within the workspace into HOCON feature config files from current context, rather than reading from python files"""
        repo_definitions = self._extract_features_from_context(anchor_list, derived_feature_list, local_workspace_dir)
        self._save_request_feature_config(repo_definitions, local_workspace_dir)
        self._save_anchored_feature_config(repo_definitions, local_workspace_dir)
        self._save_derived_feature_config(repo_definitions, local_workspace_dir)

    @classmethod
    def _save_request_feature_config(self, repo_definitions: RepoDefinitions, local_workspace_dir="./"):
        config_file_name = "feature_conf/auto_generated_request_features.conf"        
        tm = Template("""
                    // THIS FILE IS AUTO GENERATED. PLEASE DO NOT EDIT.
                    anchors: {
                        {% for anchor in feature_anchors %}
                            {% if anchor.source.name == "PASSTHROUGH" %}
                                {{anchor.to_feature_config()}}
                            {% endif %}
                        {% endfor %}
                    }
                    """)
        
        request_feature_configs = tm.render(feature_anchors=repo_definitions.feature_anchors)
        config_file_path = os.path.join(local_workspace_dir, config_file_name)
        write_to_file(content=request_feature_configs, full_file_name=config_file_path)

    @classmethod
    def _save_anchored_feature_config(self, repo_definitions: RepoDefinitions, local_workspace_dir="./"):
        config_file_name = "feature_conf/auto_generated_anchored_features.conf"
        tm = Template("""
                    // THIS FILE IS AUTO GENERATED. PLEASE DO NOT EDIT.
                    anchors: {
                        {% for anchor in feature_anchors %}
                            {% if not anchor.source.name == "PASSTHROUGH" %}
                                {{anchor.to_feature_config()}}
                            {% endif %}
                        {% endfor %}
                    }
                    
                    sources: {
                        {% for source in sources%}
                            {% if not source.name == "PASSTHROUGH" %}
                                {{source.to_feature_config()}}
                            {% endif %}
                        {% endfor %}
                    }
                    """)
        anchored_feature_configs = tm.render(feature_anchors=repo_definitions.feature_anchors,
                                             sources=repo_definitions.sources)
        config_file_path = os.path.join(local_workspace_dir, config_file_name)
        write_to_file(content=anchored_feature_configs, full_file_name=config_file_path)

    @classmethod
    def _save_derived_feature_config(self, repo_definitions: RepoDefinitions , local_workspace_dir="./"):
        config_file_name = "feature_conf/auto_generated_derived_features.conf"
        tm = Template("""
            anchors: {}
            derivations: {
                {% for derived_feature in derived_features %}
                    {{derived_feature.to_feature_config()}}
                {% endfor %}
            }
        """)
        derived_feature_configs = tm.render(derived_features=repo_definitions.derived_features)
        config_file_path = os.path.join(local_workspace_dir, config_file_name)
        write_to_file(content=derived_feature_configs, full_file_name=config_file_path)

    def register_features(self, workspace_path: Optional[Path] = None):
        """Register Features for the specified workspace

        Args:
            workspace_path (str, optional): path to a workspace. Defaults to None.
        """

        # register feature types each time when we register features.
        self._register_feathr_feature_types()
        self._read_config_from_workspace(workspace_path)
        # Upload all entities
        # need to be all in one batch to be uploaded, otherwise the GUID reference won't work
        results = self.purview_client.upload_entities(
            batch=self.entity_batch_queue)
        if results:
            webinterface_path = "https://web.purview.azure.com/resource/" + self.azure_purview_name + \
                                "/main/catalog/browseassettypes"
        else:
            RuntimeError("Feature registration failed.", results)

        logger.info(
            "Finished registering features. See {} to access the Purview web interface", webinterface_path)

    def get_registry_client(self):
        """
        Return a client object and users can operate more on it (like doing search)
        """
        return self.purview_client

    def list_registered_features(self, project_name: str = None) -> List[str]:
        """
        List all the already registered features. If project_name is not provided or is None, it will return all the
        registered features; otherwise it will only return only features under this project
        """
        entities = self.purview_client.discovery.search_entities(
            "entityType:feathr_anchor_feature")
        feature_list = []
        for entity in entities:
            # Important properties returned includes:
            # id (the guid of the entity), name, qualifiedName, @search.score,
            # and @search.highlights

            if project_name:
                # if project_name is a valid string, only append entities if the qualified name start with
                # project_name+delimiter
                if entity["qualifiedName"].startswith(project_name+self.FEATURE_REGISTRY_DELIMITER):
                    feature_list.append(entity["name"])
            else:
                # otherwise append all the entities
                feature_list.append(entity["name"])

        return feature_list
    
    def get_features_from_registry(self, project_name: str, workspace_path: str):
        """[Sync Features from registry to local workspace, given a project_name, will write project's features from registry to to user's local workspace]
        Args:
            project_name (str): project name.
            workspace_path (str): path to a workspace.
        """

        entities = self.purview_client.get_entity(qualifiedName=project_name,
                                                    typeName="feathr_workspace")
        # TODO - Change implementation to support traversing the workspace and construct the file, item by item 
        # We don't support modifying features outside of registring this should be fine.
        
        # Read the three config files from raw hocon field
        feature_conf_content = entities["entities"][0]["attributes"]["raw_hocon_feature_definition_config"]
        feature_join_conf_content = entities["entities"][0]["attributes"]["raw_hocon_feature_join_config"]
        feature_gen_conf_content = entities["entities"][0]["attributes"]["raw_hocon_feature_generation_config"]

        # Define the filenames for each config
        feature_conf_file = os.path.join(workspace_path,"feature_conf", "features.conf")
        feature_join_file = os.path.join(workspace_path,"feature_join_conf", "feature_join.conf")
        feature_gen_file = os.path.join(workspace_path,"feature_gen_conf", "feature_gen.conf")

        # Create file and directory, if does not exist
        os.makedirs(os.path.dirname(feature_conf_file), exist_ok=True)
        os.makedirs(os.path.dirname(feature_join_file), exist_ok=True)
        os.makedirs(os.path.dirname(feature_gen_file), exist_ok=True)

        with open(feature_conf_file, "w") as features:
            features.write(feature_conf_content)
        logger.info("Writing feature configuration from feathr registry to {}", feature_conf_file)
        
        with open(feature_join_file, "w") as offline_config:
            offline_config.write(feature_join_conf_content)
        logger.info("Writing offline configuration from feathr registry to {}", feature_join_file)

        with open(feature_gen_file, "w") as online_config:
            online_config.write(feature_gen_conf_content)
        logger.info("Writing online configuration from feathr registry to {}", feature_gen_file)



from setuptools import setup, find_packages
from pathlib import Path

# Use the README.md in root project
root_path = Path(__file__).resolve().parent.parent
long_description = (root_path / "README.md").read_text()

setup(
    name='feathr',
    version='0.2.0',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author_email="frame_dev@linkedin.com",
    description="An Enterprise-Grade, High Performance Feature Store",
    url="https://github.com/linkedin/feathr",
    project_urls={
        "Bug Tracker": "https://github.com/linkedin/feathr/issues",
    },
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        'Click',
        "azure-storage-file-datalake>=12.5.0",
        "azure-synapse-spark>=0.7.0",
        "azure-identity",
        "py4j",
        "loguru",
        "pandas",
        "redis",
        "requests",
        "pyapacheatlas",
        "pyhocon",
        "pandavro",
        "python-snappy",
        "pyyaml",
        "Jinja2",
        "tqdm",
        "google>=3.0.0",
        "google-api-python-client>=2.41.0",
        "pytest"
    ],
    entry_points={
        'console_scripts': ['feathr=feathrcli.cli:cli']
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.7"
)


from typing import List, Optional
from feathr.feature import Feature
from feathr.source import Source
from jinja2 import Template

# passthrough features do not need keys
class FeatureAnchor:
    """
    A feature anchor defines a set of features on top of a data source, a.k.a. a set of features anchored to a source.

    The feature producer writes multiple anchors for a feature, exposing the same feature name for the feature
    consumer to reference it.
    Attributes:
        name: Unique name of the anchor. 
        source: data source that the features are anchored to. Should be either of `INPUT_CONTEXT` or `feathr.source.Source`
        features: list of features within this anchor. 
    """
    def __init__(self,
                name: str,
                source: Source,
                features: List[Feature]):
        self.name = name
        self.features = features
        self.source = source
        self.validate_features()

    def validate_features(self):
        """Validate that anchor is non-empty and all its features share the same key"""
        assert len(self.features) > 0
        for feature in self.features:
            assert feature.key_alias == self.features[0].key_alias

    def to_feature_config(self) -> str:
        tm = Template("""
            {{anchor_name}}: {
                source: {{source.name}}
                key: [{{key_list}}]
                features: {
                    {% for feature in features %}
                        {{feature.to_feature_config()}}
                    {% endfor %}
                }
            }
        """)
        key_list = ','.join(key for key in self.features[0].key_alias)
        return tm.render(anchor_name = self.name,
                        key_list = key_list,
                        features = self.features,
                        source = self.source)



from feathrcli.cli import init
from click.testing import CliRunner
import os
from feathr.client import FeathrClient


def test_configuration_loading():
    """
    Test CLI init() is working properly.
    """
    runner = CliRunner()
    with runner.isolated_filesystem():

        result = runner.invoke(init, [])

        assert result.exit_code == 0
        assert os.path.isdir("./feathr_user_workspace")

        client = FeathrClient(config_path="./feathr_user_workspace/feathr_config.yaml")

        # test the loading is correct even if we are not in that folder
        assert client._FEATHR_JOB_JAR_PATH is not None

        TEST_LOCATION = '/test_location'

        # since we test synapse and databricks runtime using the same set of configs, we need to make sure we set both variable so that pytest can pass for both runners
        # Currently we use same runners for both the synapse and databricks runtime (so we don’t have to maintain two set of spark tests).
        # And the goal of this test is to make sure that the envs can override the yaml configs
        # In that case, we need to set both SPARK_CONFIG__AZURE_SYNAPSE__FEATHR_RUNTIME_LOCATION and SPARK_CONFIG__DATABRICKS__FEATHR_RUNTIME_LOCATION, and read the client._FEATHR_JOB_JAR_PATH variable in the code
        # client._FEATHR_JOB_JAR_PATH will read either SPARK_CONFIG__AZURE_SYNAPSE__FEATHR_RUNTIME_LOCATION or SPARK_CONFIG__DATABRICKS__FEATHR_RUNTIME_LOCATION depending on the corresponding config
        os.environ['SPARK_CONFIG__AZURE_SYNAPSE__FEATHR_RUNTIME_LOCATION'] = TEST_LOCATION
        os.environ['SPARK_CONFIG__DATABRICKS__FEATHR_RUNTIME_LOCATION'] = TEST_LOCATION

        # this should not be error out as we will just give users prompt, though the config is not really here
        client = FeathrClient(config_path="./feathr_user_workspace/feathr_config.yaml")
        assert client._FEATHR_JOB_JAR_PATH == TEST_LOCATION


import base64
import json
import os
import time
import traceback
import urllib
from collections import namedtuple
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union
from urllib.parse import urlparse
from urllib.request import urlopen

import requests
from loguru import logger
from requests.structures import CaseInsensitiveDict
from tqdm import tqdm

from feathr._abc import SparkJobLauncher
from feathr.constants import *


class _FeathrDatabricksJobLauncher(SparkJobLauncher):
    """Class to interact with Databricks Spark cluster
        This is a light-weight databricks job runner, users should use the provided template json string to get more fine controlled environment for databricks cluster.
        For example, user can control whether to use a new cluster to run the job or not, specify the cluster ID, running frequency, node size, workder no., whether to send out failed notification email, etc.
        This runner will only fill in necessary arguments in the JSON template.

        This class will read from the provided configs string, and do the following steps. 
        This default template can be overwritten by users, but users need to make sure the template is compatible with the default template. Specifically:
        1. it's a SparkJarTask (rather than other types of jobs, say NotebookTask or others). See https://docs.microsoft.com/en-us/azure/databricks/dev-tools/api/2.0/jobs#--runs-submit for more details
        2. Use the Feathr Jar to run the job (hence will add an entry in `libraries` section)
        3. Only supports `new_cluster` type for now
        4. Will override `main_class_name` and `parameters` field in the JSON template `spark_jar_task` field
        5. will override the name of this job

        Args:
            workspace_instance_url (str): the workinstance url. Document to get workspakce_instance_url: https://docs.microsoft.com/en-us/azure/databricks/workspace/workspace-details#workspace-url
            token_value (str): see here on how to get tokens: https://docs.microsoft.com/en-us/azure/databricks/dev-tools/api/latest/authentication
            config_template (str): config template for databricks cluster. See https://docs.microsoft.com/en-us/azure/databricks/dev-tools/api/2.0/jobs#--runs-submit for more details.
            databricks_work_dir (_type_, optional): databricks_work_dir must start with dbfs:/. Defaults to 'dbfs:/feathr_jobs'.
        """  
    def __init__(
            self,
            workspace_instance_url: str,
            token_value: str,
            config_template: Union[str,Dict],
            databricks_work_dir: str = 'dbfs:/feathr_jobs',
    ):
          

        # Below we will use Databricks job APIs (as well as many other APIs) to submit jobs or transfer files
        # For Job APIs, see https://docs.microsoft.com/en-us/azure/databricks/dev-tools/api/2.0/jobs
        # for DBFS APIs, see: https://docs.microsoft.com/en-us/azure/databricks/dev-tools/api/latest/dbfs
        self.config_template = config_template
        # remove possible trailing '/' due to wrong input format
        self.workspace_instance_url = workspace_instance_url.rstrip('/')
        self.auth_headers = CaseInsensitiveDict()

        # Authenticate the REST APIs. Documentation: https://docs.microsoft.com/en-us/azure/databricks/dev-tools/api/latest/authentication
        self.auth_headers['Accept'] = 'application/json'
        self.auth_headers['Authorization'] = f'Bearer {token_value}'
        self.databricks_work_dir = databricks_work_dir

    def upload_or_get_cloud_path(self, local_path_or_http_path: str):
        """
        Supports transferring file from an http path to cloud working storage, or upload directly from a local storage.
        """
        src_parse_result = urlparse(local_path_or_http_path)
        file_name = os.path.basename(local_path_or_http_path)
        # returned paths for the uploaded file
        returned_path = os.path.join(self.databricks_work_dir, file_name)
        if src_parse_result.scheme.startswith('http'):
            with urlopen(local_path_or_http_path) as f:
                data = f.read()
                files = {'file': data}
                # for DBFS APIs, see: https://docs.microsoft.com/en-us/azure/databricks/dev-tools/api/latest/dbfs
                r = requests.post(url=self.workspace_instance_url+'/api/2.0/dbfs/put',
                                  headers=self.auth_headers, files=files,  data={'overwrite': 'true', 'path': returned_path})
                logger.debug('{} is downloaded and then uploaded to location: {}',
                             local_path_or_http_path, returned_path)
        elif src_parse_result.scheme.startswith('dbfs'):
            # passed a cloud path
            logger.debug(
                'Skipping file {} as it is already in the cloud', local_path_or_http_path)
            returned_path = local_path_or_http_path
        else:
            # else it should be a local file path or dir
            if os.path.isdir(local_path_or_http_path):
                logger.info("Uploading folder {}", local_path_or_http_path)
                dest_paths = []
                for item in Path(local_path_or_http_path).glob('**/*.conf'):
                    returned_path = self.upload_file(item.resolve())
                    dest_paths.extend([returned_path])
                returned_path = ','.join(dest_paths)
            else:
                returned_path = self.upload_file(local_path_or_http_path)
        return returned_path

    def upload_file(self, local_path_or_http_path: str) -> str:
        """
        Supports transferring file from an http path to cloud working storage, or upload directly from a local storage.
        """
        file_name = os.path.basename(local_path_or_http_path)
        # returned paths for the uploaded file
        returned_path = os.path.join(self.databricks_work_dir, file_name)
        with open(local_path_or_http_path, 'rb') as f:
            data = f.read()
            files = {'file': data}
            # for DBFS APIs, see: https://docs.microsoft.com/en-us/azure/databricks/dev-tools/api/latest/dbfs
            r = requests.post(url=self.workspace_instance_url+'/api/2.0/dbfs/put',
                              headers=self.auth_headers, files=files,  data={'overwrite': 'true', 'path': returned_path})
            logger.debug('{} is uploaded to location: {}',
                         local_path_or_http_path, returned_path)
        return returned_path

    def submit_feathr_job(self, job_name: str, main_jar_path: str,  main_class_name: str, arguments: List[str], reference_files_path: List[str] = [], job_tags: Dict[str, str] = None, configuration: Dict[str, str] = None):
        """
        submit the feathr job to databricks
        Refer to the databricks doc for more details on the meaning of the parameters:
        https://docs.microsoft.com/en-us/azure/databricks/dev-tools/api/2.0/jobs#--runs-submit
        Args:
            main_file_path (str): main file paths, usually your main jar file
            main_class_name (str): name of your main class
            arguments (str): all the arugments you want to pass into the spark job
            job_tags (str): tags of the job, for exmaple you might want to put your user ID, or a tag with a certain information
            configuration (Dict[str, str]): Additional configs for the spark job
        """

        
        if isinstance(self.config_template, str):
            # if the input is a string, load it directly
            submission_params = json.loads(self.config_template)
        else:
            # otherwise users might have missed the quotes in the config. 
            submission_params = self.config_template
            logger.warning("Databricks config template loaded in a non-string fashion. Please consider providing the config template in a string fashion.")
        
        submission_params['run_name'] = job_name
        submission_params['libraries'][0]['jar'] = self.upload_or_get_cloud_path(main_jar_path)
        submission_params['new_cluster']['spark_conf'] = configuration
        submission_params['new_cluster']['custom_tags'] = job_tags
        submission_params['spark_jar_task']['parameters'] = arguments
        submission_params['spark_jar_task']['main_class_name'] = main_class_name
        self.res_job_id = None

        try:
            # For Job APIs, see https://docs.microsoft.com/en-us/azure/databricks/dev-tools/api/2.0/jobs
            result = requests.post(url=self.workspace_instance_url+'/api/2.0/jobs/runs/submit',
                              headers=self.auth_headers, data=json.dumps(submission_params))
            self.res_job_id = result.json()['run_id']
            # For Job APIs, see https://docs.microsoft.com/en-us/azure/databricks/dev-tools/api/2.0/jobs
            result = requests.get(url=self.workspace_instance_url+'/api/2.0/jobs/runs/get',
                             headers=self.auth_headers, params={'run_id': str(self.res_job_id)})
            logger.info('Feathr Job Submitted Sucessfully. View more details here: {}', result.json()[
                        'run_page_url'])
        except:
            traceback.print_exc()
        # return ID as the submission result
        return self.res_job_id

    def wait_for_completion(self, timeout_seconds: Optional[int] = 500) -> bool:
        """ Returns true if the job completed successfully
        """
        start_time = time.time()
        while (timeout_seconds is None) or (time.time() - start_time < timeout_seconds):
            status = self.get_status()
            logger.debug('Current Spark job status: {}', status)
            # see all the status here:
            # https://docs.microsoft.com/en-us/azure/databricks/dev-tools/api/2.0/jobs#--runlifecyclestate
            # https://docs.microsoft.com/en-us/azure/databricks/dev-tools/api/2.0/jobs#--runresultstate
            if status in {'SUCCESS'}:
                return True
            elif status in {'INTERNAL_ERROR', 'FAILED', 'TIMEDOUT', 'CANCELED'}:
                return False
            else:
                time.sleep(30)
        else:
            raise TimeoutError('Timeout waiting for job to complete')

    def get_status(self) -> str:
        assert self.res_job_id is not None
        # For Job Runs APIs, see https://docs.microsoft.com/en-us/azure/databricks/dev-tools/api/2.0/jobs#--runs-get
        result = requests.get(url=self.workspace_instance_url+'/api/2.0/jobs/runs/get',
                               headers=self.auth_headers, params={'run_id': str(self.res_job_id)})
        # first try to get result state. it might not be available, and if that's the case, try to get life_cycle_state
        res_state = result.json()['state'].get('result_state') or result.json()[
            'state']['life_cycle_state']
        assert res_state is not None
        return res_state
        
    def get_job_result_uri(self) -> str:
        """Get job output uri

        Returns:
            str: `output_path` field in the job tags
        """
        assert self.res_job_id is not None
        # For Job Runs APIs, see https://docs.microsoft.com/en-us/azure/databricks/dev-tools/api/2.0/jobs#--runs-get
        result = requests.get(url=self.workspace_instance_url+'/api/2.0/jobs/runs/get',
                               headers=self.auth_headers, params={'run_id': str(self.res_job_id)})
        custom_tags = result.json()['cluster_spec']['new_cluster']['custom_tags']
        assert custom_tags is not None
        return custom_tags[OUTPUT_PATH_TAG]

    def download_result(self, result_path: str, local_folder: str):
        """
        Supports downloading files from the result folder. Only support paths starts with `dbfs:/` and only support downloading files in one folder (per Spark's design, everything will be in the result folder in a flat manner)
        """
        if not result_path.startswith('dfbs'):
            RuntimeError('Currently only paths starting with dbfs is supported for downloading results from a databricks cluster. The path should start with \"dbfs:\" .')

        try:
            # listing all the files in a folder: https://docs.microsoft.com/en-us/azure/databricks/dev-tools/api/latest/dbfs#--list
            result = requests.get(url=self.workspace_instance_url+'/api/2.0/dbfs/list',
                                headers=self.auth_headers,  params={ 'path': result_path})
            dbfs_files = result.json()['files']
            for file_path in tqdm(dbfs_files, desc="Downloading result files: "):
                # each file_path would be a dict of this type: https://docs.microsoft.com/en-us/azure/databricks/dev-tools/api/latest/dbfs#dbfsfileinfo
                dbfs_file_path, dbfs_file_size, local_file_path = file_path['path'], file_path['file_size'], os.path.join(local_folder, os.path.basename(file_path['path']))
                with open(local_file_path, 'wb') as file_obj:
                    downloaded_size = 0
                    # Loop until we've downloaded the whole file
                    while downloaded_size < dbfs_file_size:
                        chunk = self._read_single_chunk(path=dbfs_file_path, offset=downloaded_size, length=MB_BYTES)
                        file_obj.write(base64.b64decode(chunk.data))
                        downloaded_size += chunk.bytes_read
            # Logging once all the download is finished.
            logger.info('Finish downloading files from {} to {}.', result_path,local_folder)
        except requests.exceptions.RequestException as e:  # This is the correct syntax
            raise SystemExit(e)



    def _read_single_chunk(self, path, offset, length=MB_BYTES):

        params = {"path": path,
                "offset": offset,
                "length": length}

        # get a single chunk of file
        # https://docs.microsoft.com/en-us/azure/databricks/dev-tools/api/latest/dbfs#--read
        resp = requests.get(url=self.workspace_instance_url+'/api/2.0/dbfs/read',  headers=self.auth_headers,  params=params)
        FileReadInfo = namedtuple("FileReadInfo", ['bytes_read', 'data'])
        if resp.status_code == 200:
            return FileReadInfo(**resp.json())
        else:
            RuntimeError("Files cannot be downloaded.")




from abc import ABC, abstractmethod
from typing import List, Optional
from jinja2 import Template


class Source:
    """External data source for feature. Typically a 'table'.
     Attributes:
         name: name of the source
         event_timestamp_column: column name of the event timestamp
         timestamp_format: the format of the event_timestamp_column, e.g. yyyy/MM/DD.
    """
    def __init__(self,
                 name: str,
                 event_timestamp_column: Optional[str], 
                 timestamp_format: Optional[str] = "epoch") -> None:
        self.name = name
        self.event_timestamp_column = event_timestamp_column
        self.timestamp_format = timestamp_format

    def __eq__(self, other):
        """A source is equal to another if name is equal."""
        return self.name == other.name

    def __hash__(self):
        """A source can be identified with the name"""
        return hash(self.name)

    def to_write_config(self) -> str:
        pass

class InputContext(Source):
    """A type of 'passthrough' source, a.k.a. request feature source.
    """
    SOURCE_NAME = "PASSTHROUGH"
    def __init__(self) -> None:
        super().__init__(self.SOURCE_NAME, None, None)

    def to_feature_config(self) -> str:
        return "source: " + self.name


class HdfsSource(Source):
    """A data source(table) stored on HDFS-like file system. Data can be fetch through a POSIX style path.

        Args:
            name (str): name of the source
            path (str): The location of the source data. 
            event_timestamp_column (Optional[str]): The timestamp field of your record. As sliding window aggregation feature assume each record in the source data should have a timestamp column. 
            timestamp_format (Optional[str], optional): The format of the timestamp field. Defaults to "epoch". Possible values are: 
            - `epoch` (seconds since epoch), for example `1647737463`
            - `epoch_millis` (milliseconds since epoch), for example `1647737517761`
            - Any date formats supported by [SimpleDateFormat](https://docs.oracle.com/javase/8/docs/api/java/text/SimpleDateFormat.html). 
                
        """
    def __init__(self, name: str, path: str, event_timestamp_column: Optional[str]= None, timestamp_format: Optional[str] = "epoch") -> None:
        super().__init__(name, event_timestamp_column, timestamp_format)
        self.path = path

    def to_feature_config(self) -> str:
        tm = Template("""  
            {{source.name}}: {
                location: {path: "{{source.path}}"}
                {% if source.event_timestamp_column is defined %}
                    timeWindowParameters: {
                        timestampColumn: "{{source.event_timestamp_column}}"
                        timestampColumnFormat: "{{source.timestamp_format}}"
                    }
                {% endif %}
            } 
        """)
        msg = tm.render(source=self)
        return msg
 
INPUT_CONTEXT = InputContext()


import os
from pathlib import Path


def write_to_file(content: str, full_file_name: str):
    """Write content to a file.
    Attributes:
        content: content to write into the file
        full_file_name: full file path
    """
    file_name_start = full_file_name.rfind("/")
    if file_name_start > 0:
        dir_name = full_file_name[:file_name_start]
        Path(dir_name).mkdir(parents=True, exist_ok=True)
    with open(full_file_name, "w") as handle:
        print(content, file=handle)